"""
Three-Point Clustering Statistics (:mod:`~triumvirate.threept`)
===============================================================

Measure three-point clustering statistics from catalogues.

Local plane-parallel estimators
-------------------------------

.. autosummary::
    compute_bispec
    compute_3pcf

Global plane-parallel estimators
--------------------------------

.. autosummary::
    compute_bispec_in_gpp_box
    compute_3pcf_in_gpp_box

Window function estimator
-------------------------

.. autosummary::
    compute_3pcf_window

"""  # numpydoc ignore=GL07,SS01
import warnings
from pathlib import Path

import numpy as np

from ._threept import (
    _calc_bispec_normalisation_from_mesh,
    _calc_bispec_normalisation_from_particles,
    _compute_bispec,
    # _compute_bispec_for_los_choice,
    _compute_bispec_in_gpp_box,
    _compute_3pcf,
    _compute_3pcf_in_gpp_box,
    _compute_3pcf_window,
)
from .dataobjs import Binning
from .parameters import (
    _modify_measurement_parameters,
    _modify_sampling_parameters,
    fetch_paramset_template,
    ParameterSet,
)


def _amalgamate_parameters(paramset=None, params_sampling=None,
                           degrees=None, wa_orders=None, binning=None,
                           form=None, idx_bin=None, **type_kwargs):
    """Amalgamate a parameter set by overriding sampling parameters and
    measurement parameters.

    Parameters
    ----------
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set.  If `None` (default), some other parameters
        should be provided (see below).
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    wa_orders : tuple of (int, int) or str of length 2, optional
        Wide-angle correction orders either as a tuple ('i_wa', 'j_wa') or
        as a string of length 2.  If not `None` (default), this will
        override ``paramset['wa_orders']`` entries.  If a string,
        multipole degrees are assumed to be single-digit integers.
    binning : :class:`~.triumvirate.dataobjs.Binning`, optional
        Binning (default is `None`).
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    **type_kwargs
        'catalogue_type' and 'statistic_type' parameters to be filled in.

    Returns
    -------
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Amalgamated full parameter set.

    Raises
    ------
    ValueError
        When `paramset` is `None` while `params_sampling`, `degrees`,
        `binning` or `form` is also `None`, or while `form` is 'full' and
        `idx_bin` is `None`.

    """
    if paramset is None and params_sampling is None:
        raise ValueError(
            "Either `paramset` or `sampling_params` must be provided."
        )
    if paramset is None and None in [degrees, binning, form]:
        raise ValueError(
            "`degrees`, `binning` and `form` must be provided "
            "when `paramset` is None."
        )
    if paramset is None and form in ['row', 'off-diag'] and idx_bin is None:
        raise ValueError(
            "`idx_bin` must be provided when `paramset` is None "
            "and `form` is 'row' or 'off-diag'."
        )

    if paramset is None:
        paramset, defaults = fetch_paramset_template('dict', ret_defaults=True)
    else:
        # paramset = paramset
        defaults = None

    if params_sampling is not None:
        paramset, defaults = _modify_sampling_parameters(
            paramset, params_sampling,
            params_default=defaults, ret_defaults=True
        )

    params_measure = {}
    if degrees is not None:
        ell1, ell2, ELL = degrees
        params_measure.update({
            'ell1': int(ell1), 'ell2': int(ell2), 'ELL': int(ELL)
        })
    if wa_orders is not None:
        i_wa, j_wa = wa_orders
        params_measure.update({
            'i_wa': int(i_wa), 'j_wa': int(j_wa)
        })
    if binning is not None:
        params_measure.update({
            'bin_min': binning.bin_min,
            'bin_max': binning.bin_max,
            'num_bins': binning.num_bins,
        })
    if form is not None:
        params_measure.update({'form': form})
    if idx_bin is not None:
        params_measure.update({'idx_bin': idx_bin})

    paramset, defaults = _modify_measurement_parameters(
        paramset, params_measure,
        params_default=defaults, ret_defaults=True
    )

    # Implicitly validate parameters.
    paramset.update(type_kwargs)

    if defaults:
        warnings.warn(
            "The following parameter default values "
            f"are unchanged: {defaults}. "
            "Not all parameters are necessarily used."
        )

    return paramset


def _get_measurement_filename(paramset):
    """Get output measurement filename.

    Parameters
    ----------
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set.

    Returns
    -------
    str
        Output measurement filename.

    Raises
    ------
    ValueError
        When `paramset` indicates the measurements are not
        three-point statistics.

    """
    multipole = "{ell1}{ell2}{ELL}".format(**paramset['degrees'])

    if paramset['form'] == 'diag':
        binform = "_diag"
    if paramset['form'] == 'off-diag':
        binform = "_offdiag{:d}".format(paramset['idx_bin'])
    if paramset['form'] == 'row':
        binform = "_row{:d}".format(paramset['idx_bin'])
    if paramset['form'] == 'full':
        binform = "_full"

    output_tag = paramset['tags']['output'] or ""

    if paramset['statistic_type'] == 'bispec':
        return "bk{}{}{}".format(multipole, binform, output_tag)
    if paramset['statistic_type'] == '3pcf':
        return "zeta{}{}{}".format(multipole, binform, output_tag)
    if paramset['statistic_type'] == '3pcf-win':
        return "zetaw{}{}{}".format(multipole, binform, output_tag)
    if paramset['statistic_type'] == '3pcf-win-wa':
        wa_tag = "_wa{i}{j}".format(**paramset['wa_orders'])
        return "zetaw{}{}{}{}".format(multipole, binform, wa_tag, output_tag)

    raise ValueError(
        "`paramset` 'statistic_type' does not correspond to a "
        "recognised three-point statistic."
    )


def _print_measurement_header(paramset, norm_factor_part, norm_factor_mesh,
                              norm_factor_meshes):
    """Print three-point statistic measurement header including sampling
    parameters, normalisation factors and data table columns.

    Parameters
    ----------
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set.
    norm_factor_part, norm_factor_mesh, norm_factor_meshes : float
        Normalisation factors.

    Returns
    -------
    text_header : str
        Measurement information as a header string.

    Raises
    ------
    ValueError
        When `paramset` indicates the measurements are not
        three-point statistics.

    """
    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
    if paramset['norm_convention'] == 'mesh-mixed':
        norm_factor = norm_factor_meshes

    if paramset['npoint'] != '3pt':
        raise ValueError(
            "`paramset` 'statistic_type' does not correspond to a "
            "recognised three-point statistic."
        )

    multipole = "{ell1}{ell2}{ELL}".format(**paramset['degrees'])
    if paramset['space'] == 'fourier':
        datatab_colnames = [
            "k1_cen", "k1_eff", "nmodes_1", "k2_cen", "k2_eff", "nmodes_2",
            "Re{{bk{}_raw}}".format(multipole),
            "Im{{bk{}_raw}}".format(multipole),
            "Re{{bk{}_shot}}".format(multipole),
            "Im{{bk{}_shot}}".format(multipole)
        ]
    if paramset['space'] == 'config':
        datatab_colnames = [
            "r1_cen", "r1_eff", "npairs_1", "r2_cen", "r2_eff", "npairs_2",
            "Re{{zeta{}_raw}}".format(multipole),
            "Im{{zeta{}_raw}}".format(multipole),
            "Re{{zeta{}_shot}}".format(multipole),
            "Im{{zeta{}_shot}}".format(multipole)
        ]

    text_lines = [
        "Box size: [{:.3f}, {:.3f}, {:.3f}]".format(
            *[paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
        ),
        "Box alignment: {}".format(paramset['alignment']),
        "Mesh number: [{:d}, {:d}, {:d}]".format(
            *[paramset['ngrid'][ax] for ax in ['x', 'y', 'z']]
        ),
        "Mesh assignment and interlacing: {}, {}".format(
            paramset['assignment'], paramset['interlace']
        ),
        "Normalisation factor: {:.9e} ({})".format(
            norm_factor, paramset['norm_convention']
        ),
        "Normalisation factor alternatives: "
        "{:.9e} (particle), {:.9e} (mesh), {:.9e} (mesh-mixed; n/a)".format(
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        ),
        ", ".join([
            "[{:d}] {}".format(colidx, colname)
            for colidx, colname in enumerate(datatab_colnames)
        ])
    ]

    text_header = "\n".join(text_lines)

    return text_header


def _assemble_measurement_datatab(measurements, paramset):
    """Assemble measurement data table.

    Parameters
    ----------
    measurements : dict
        Measurement results.
    paramset : :class:`~triumvirate.parameters.ParameterSet`
        Parameter set.

    Returns
    -------
    datatab : :class:`numpy.ndarray`
        Column-major measurement data table.

    Raises
    ------
    ValueError
        When `paramset` indicates the measurements are not
        three-point statistics.

    """
    if paramset['npoint'] != '3pt':
        raise ValueError(
            "`paramset` 'statistic_type' does not correspond to a "
            "recognised three-point statistic."
        )
    if paramset['space'] == 'fourier':
        datatab = np.transpose([
            measurements['k1_bin'], measurements['k1_eff'],
            measurements['nmodes_1'],
            measurements['k2_bin'], measurements['k2_eff'],
            measurements['nmodes_2'],
            measurements['bk_raw'].real, measurements['bk_raw'].imag,
            measurements['bk_shot'].real, measurements['bk_shot'].imag,
        ])
    if paramset['space'] == 'config':
        datatab = np.transpose([
            measurements['r1_bin'], measurements['r1_eff'],
            measurements['npairs_1'],
            measurements['r2_bin'], measurements['r2_eff'],
            measurements['npairs_2'],
            measurements['zeta_raw'].real, measurements['zeta_raw'].imag,
            measurements['zeta_shot'].real, measurements['zeta_shot'].imag,
        ])

    return datatab


# ========================================================================
# Survey statistics
# ========================================================================

def _compute_3pt_stats_survey_like(threept_algofunc,
                                   catalogue_data, catalogue_rand,
                                   los_data=None, los_rand=None,
                                   paramset=None, params_sampling=None,
                                   degrees=None, binning=None,
                                   form=None, idx_bin=None, types=None,
                                   save=False, logger=None):
    """Compute three-point statistics from survey-like data and random
    catalogues in the local plane-parallel approximation.

    Parameters
    ----------
    threept_algofunc : callable
        Three-point statistic algorithmic function.
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_data : (N, 3) array of float, optional
        Specified lines of sight for the data-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set.  If `None` (default), `degrees`, `binning`,
        `form` and `params_sampling` should be provided; `idx_bin`
        should be provided when `form` is 'full'.
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    types : dict, optional
        'catalogue_type' and 'statistic_type' values (default is `None`).
        This should be set by the caller of this function.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    """
    # --------------------------------------------------------------------
    # Initialisation
    # --------------------------------------------------------------------

    # -- Parameters ------------------------------------------------------

    paramset = _amalgamate_parameters(
        paramset=paramset, params_sampling=params_sampling,
        degrees=degrees, binning=binning, form=form, idx_bin=idx_bin, **types
    )

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset, logger=logger)

    if logger:
        logger.info("Parameter set have been initialised.")

    if paramset['catalogue_type'] != 'survey':
        raise ValueError(
            "`paramset` 'catalogue_type' does not correspond to "
            "the local plane-parallel clustering algorithm being called: "
            f"catalogue_type = '{paramset['catalogue_type']}'."
        )

    if paramset['statistic_type'] == 'bispec':
        statistic_name = 'bispectrum'
    elif paramset['statistic_type'] == '3pcf':
        statistic_name = 'three-point correlation function'
    else:
        raise ValueError(
            "`paramset` 'statistic_type' does not correspond to "
            "the clustering algorithm being called: "
            f"statistic_type = '{paramset['statistic_type']}'."
        )

    # -- Data ------------------------------------------------------------

    # Set up binning.
    if binning is None:
        binning = Binning.from_parameter_set(paramset)

    if logger:
        logger.info("Binning has been initialised.")

    # Set up lines of sight.
    if los_data is None:
        los_data = catalogue_data.compute_los()
    if los_rand is None:
        los_rand = catalogue_rand.compute_los()

    los_data = np.ascontiguousarray(los_data)
    los_rand = np.ascontiguousarray(los_rand)

    if logger:
        logger.info("Lines of sight have been initialised.")

    # Set up box alignment.
    if paramset['alignment'] == 'centre':
        catalogue_data.centre(
            [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']],
            catalogue_ref=catalogue_rand
        )
    if paramset['alignment'] == 'pad':
        if paramset['padscale'] == 'box':
            kwargs = {'boxsize_pad': paramset['padfactor']}
        if paramset['padscale'] == 'grid':
            kwargs = {
                'ngrid': [paramset['ngrid'][ax] for ax in ['x', 'y', 'z']],
                'ngrid_pad': paramset['padfactor']
            }
        catalogue_data.pad(
            [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']],
            catalogue_ref=catalogue_rand, **kwargs
        )

    if logger:
        logger.info("Catalogues have been aligned.")

    # --------------------------------------------------------------------
    # Measurements
    # --------------------------------------------------------------------

    # Prepare catalogues.
    if logger:
        logger.info(
            "Preparing catalogue for clustering algorithm...",
            cpp_state='start'
        )

    particles_data = \
        catalogue_data._convert_to_cpp_catalogue(verbose=paramset['verbose'])
    particles_rand = \
        catalogue_rand._convert_to_cpp_catalogue(verbose=paramset['verbose'])

    if logger:
        logger.info(
            "... prepared catalogue for clustering algorithm.",
            cpp_state='end'
        )

    # Set up constants.
    alpha = catalogue_data.wstotal / catalogue_rand.wstotal

    if logger:
        logger.info("Alpha contrast: %.6e.", alpha)

    norm_factor_part = _calc_bispec_normalisation_from_particles(
        particles_rand, alpha
    )
    norm_factor_mesh = _calc_bispec_normalisation_from_mesh(
        particles_rand, paramset, alpha
    )
    norm_factor_meshes = 0.

    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; n/a) (none used)."
        )
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed; n/a)."
        )
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed; n/a)."
        )
    if paramset['norm_convention'] == 'mesh-mixed':
        norm_factor = norm_factor_meshes
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; used)."
        )

    if logger:
        logger.info(
            norm_log_mesg,
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        )

    # Perform measurement.
    if logger:
        logger.info(
            "Measuring %s from paired survey-type catalogues "
            "in the local plane-parallel approximation...",
            statistic_name,
            cpp_state='start'
        )

    results = threept_algofunc(
        particles_data, particles_rand, los_data, los_rand,
        paramset, binning, norm_factor
    )

    if logger:
        logger.info(
            "... measured %s from paired survey-type catalogues "
            "in the local plane-parallel approximation.",
            statistic_name,
            cpp_state='end'
        )

    if save:
        odirpath = paramset['directories']['measurements'] or ""
        header = "\n".join([
            catalogue_data.write_attrs_as_header(catalogue_ref=catalogue_rand),
            _print_measurement_header(
                paramset,
                norm_factor_part, norm_factor_mesh, norm_factor_meshes
            ),
        ])
        if save.lower() == '.txt':
            datatab = _assemble_measurement_datatab(results, paramset)
            datafmt = '\t'.join(
                (['%.9e'] * 2 + ['%10d']) * 2 +
                ['% .9e'] * (datatab.shape[-1] - 6)
            )
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
            np.savetxt(
                ofilepath, datatab, fmt=datafmt, header=header, delimiter='\t'
            )
        elif save.lower().endswith('.npz'):
            results.update({'header': header})
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
            np.savez(ofilepath, **results)
        else:
            raise ValueError(
                f"Unrecognised save format for measurements: {save}."
            )

        if logger:
            logger.info("Measurements saved to %s.", ofilepath)

    return results


def compute_bispec(catalogue_data, catalogue_rand,
                   los_data=None, los_rand=None,
                   degrees=None, binning=None, form=None, idx_bin=None,
                   sampling_params=None,
                   paramset=None,
                   save=False, logger=None):
    """Compute bispectrum from survey-like data and random catalogues in
    the local plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_data : (N, 3) array of float, optional
        Specified lines of sight for the data-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used in lieu of
        `degrees`, `binning`, `form`, `idx_bin` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results as a dictionary with the following entries---

        - 'k1_bin', 'k2_bin': central wavenumber for each bin of
          the first and second wavenumbers;
        - 'k1_eff', 'k2_eff': effective wavenumber for each bin of
          the first and second wavenumbers;
        - 'nmodes_1', 'nmodes_2': number of wavevector modes in each bin
          of the first and second wavenumbers;
        - 'bk_raw': bispectrum raw measurements including any
          specified normalisation and shot noise;
        - 'bk_shot': bispectrum shot noise.

        The effective wavenumber is here defined as the average wavenumber
        in each bin.

    Examples
    --------
    Specify multipole `degrees` (0, 0, 2) and bispectrum 'row' `form` with
    the first wavenumber fixed in the bin indexed 5.  These override the
    corresponding parameters supplied by `paramset`.

    >>> results = compute_bispec(
    ...     catalogue_data, catalogue_rand,
    ...     degrees=(0, 0, 2),
    ...     form='full',
    ...     idx_bin=5,
    ...     paramset=paramset
    ... )

    See more analogous examples in
    :func:`~triumvirate.twopt.compute_powspec`.

    """
    results = _compute_3pt_stats_survey_like(
        _compute_bispec,
        catalogue_data, catalogue_rand,
        los_data=los_data, los_rand=los_rand,
        paramset=paramset, params_sampling=sampling_params,
        degrees=degrees, binning=binning, form=form, idx_bin=idx_bin,
        types={'catalogue_type': 'survey', 'statistic_type': 'bispec'},
        save=save, logger=logger
    )

    return results


def compute_3pcf(catalogue_data, catalogue_rand,
                 los_data=None, los_rand=None,
                 degrees=None, binning=None, form=None, idx_bin=None,
                 sampling_params=None,
                 paramset=None,
                 save=False, logger=None):
    """Compute three-point correlation function from survey-like data
    and random catalogues in the local plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_data : (N, 3) array of float, optional
        Specified lines of sight for the data-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used in lieu of
        `degrees`, `binning`, `form`, `idx_bin` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results as a dictionary with the following entries---

        - 'r1_bin', 'r2_bin': central separation for each bin of
          the first and second separations;
        - 'r1_eff', 'r2_eff': effective separation for each bin of
          the first and second separations;
        - 'npairs_1', 'npairs_2': number of separation pairs in each bin
          of the first and second separations;
        - 'zeta_raw': three-point correlation function raw measurements
          including any specified normalisation and shot noise;
        - 'zeta_shot': three-point correlation function shot noise.

        The effective separation is here defined as the average separation
        in each bin.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.threept.compute_bispec`.

    """
    results = _compute_3pt_stats_survey_like(
        _compute_3pcf,
        catalogue_data, catalogue_rand,
        los_data=los_data, los_rand=los_rand,
        paramset=paramset, params_sampling=sampling_params,
        degrees=degrees, binning=binning, form=form, idx_bin=idx_bin,
        types={'catalogue_type': 'survey', 'statistic_type': '3pcf'},
        save=save, logger=logger
    )

    return results

# def compute_bispec_for_los_choice(catalogue_data, catalogue_rand, los_choice,
#                                   los_data=None, los_rand=None,
#                                   degrees=None, binning=None,
#                                   form=None, idx_bin=None,
#                                   sampling_params=None,
#                                   paramset=None,
#                                   save=False, logger=None):
#     """Compute bispectrum from data and random catalogues.
#
#     Parameters
#     ----------
#     catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
#         Data-source catalogue.
#     catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
#         Random-source catalogue.
#     los_choice : {0, 1, 2}, int
#         Line-of-sight end-point choice.
#     los_data : (N, 3) array of float, optional
#         Specified lines of sight for the data-source catalogue.
#         If `None` (default), this is automatically computed using
#         :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
#     los_rand : (N, 3) array of float, optional
#         Specified lines of sight for the random-source catalogue.
#         If `None` (default), this is automatically computed using
#         :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
#     degrees : tuple of (int, int, int) or str of length 3, optional
#         Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
#         as a string of length 3.  If not `None` (default), this will
#         override ``paramset['degrees']`` entries.  If a string, multipole
#         degrees are assumed to be single-digit integers.
#     binning : :class:`~triumvirate.dataobjs.Binning`, optional
#         Binning for the measurements.  If `None` (default),
#         this is constructed from `paramset`.
#     form : {'diag', 'off-diag', 'row', 'full'}, optional
#         Binning form of the measurements.  If not `None` (default),
#         this will override ``paramset['form']``.
#
#         .. versionchanged:: 0.3.0
#             Add 'off-diag' and 'row' options and redefine 'full' option.
#
#     idx_bin : int, optional
#         When binning `form` is 'row', this is the fixed bin index for
#         the first coordinate dimension; when binning `form` is 'off-diag',
#         this is the upper-triangular off-diagonal index; otherwise, this
#         is ignored.  If not `None` (default), this will override
#         ``paramset['idx_bin']``.
#
#         .. versionchanged:: 0.3.0
#             Expand definition to match changed `form`.
#
#     sampling_params : dict, optional
#         Dictionary containing a subset of the following entries
#         for sampling parameters---
#
#         - 'alignment': {'centre', 'pad'};
#         - 'boxsize': sequence of [float, float, float];
#         - 'ngrid': sequence of [int, int, int];
#         - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
#         - 'interlace': bool;
#
#         and exactly one of the following only when 'alignment' is 'pad'---
#
#         - 'boxpad': float;
#         - 'gridpad': float.
#
#         If not `None` (default), this will override the corresponding
#         entries in `paramset`.
#     paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
#         Full parameter set (default is `None`).  This is used in lieu of
#         `degrees`, `binning`, `form`, `idx_bin` or `sampling_params`.
#     save : {'.txt', '.npz', False}, optional
#         If not `False` (default), save the measurements as a '.txt' file
#         or in '.npz' format.  The save path is determined from `paramset`
#         (if unset, a default file path in the current working directory is
#         used).
#     logger : :class:`logging.Logger`, optional
#         Logger (default is `None`).
#
#     Returns
#     -------
#     results : dict of {str: :class:`numpy.ndarray`}
#         Measurement results as a dictionary with the following entries---
#
#         - 'k1_bin', 'k2_bin': central wavenumber for each bin of
#           the first and second wavenumbers;
#         - 'k1_eff', 'k2_eff': effective wavenumber for each bin of
#           the first and second wavenumbers;
#         - 'nmodes_1', 'nmodes_2': number of wavevector modes in each bin
#           of the first and second wavenumbers;
#         - 'bk_raw': bispectrum raw measurements including any
#           specified normalisation and shot noise;
#         - 'bk_shot': bispectrum shot noise.
#
#         The effective wavenumber is here defined as the average wavenumber
#         in each bin.
#
#     """
#     # --------------------------------------------------------------------
#     # Initialisation
#     # --------------------------------------------------------------------

#     # -- Parameters ------------------------------------------------------

#     paramset = _amalgamate_parameters(
#         paramset=paramset, params_sampling=sampling_params,
#         degrees=degrees, binning=binning, form=form, idx_bin=idx_bin,
#         catalogue_type='survey', statistic_type='bispec'
#     )

#     if isinstance(paramset, dict):  # likely redundant but safe
#         paramset = ParameterSet(param_dict=paramset, logger=logger)

#     if logger:
#         logger.info("Parameter set have been initialised.")

#     # -- Data ------------------------------------------------------------

#     # Set up binning.
#     if binning is None:
#         binning = Binning.from_parameter_set(paramset)

#     if logger:
#         logger.info("Binning has been initialised.")

#     # Set up lines of sight.
#     if los_data is None:
#         los_data = catalogue_data.compute_los()
#     if los_rand is None:
#         los_rand = catalogue_rand.compute_los()

#     los_data = np.ascontiguousarray(los_data)
#     los_rand = np.ascontiguousarray(los_rand)

#     if logger:
#         logger.info("Lines of sight have been initialised.")

#     # Set up box alignment.
#     if paramset['alignment'] == 'centre':
#         catalogue_data.centre(
#             [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']],
#             catalogue_ref=catalogue_rand
#         )
#     if paramset['alignment'] == 'pad':
#         if paramset['padscale'] == 'box':
#             kwargs = {'boxsize_pad': paramset['padfactor']}
#         if paramset['padscale'] == 'grid':
#             kwargs = {
#                 'ngrid': [paramset['ngrid'][ax] for ax in ['x', 'y', 'z']],
#                 'ngrid_pad': paramset['padfactor']
#             }
#         catalogue_data.pad(
#             [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']],
#             catalogue_ref=catalogue_rand, **kwargs
#         )

#     if logger:
#         logger.info("Catalogues have been aligned.")

#     # --------------------------------------------------------------------
#     # Measurements
#     # --------------------------------------------------------------------

#     # Prepare catalogues.
#     if logger:
#         logger.info(
#             "Preparing catalogue for clustering algorithm...",
#             cpp_state='start'
#         )

#     particles_data = \
#         catalogue_data._convert_to_cpp_catalogue(verbose=paramset['verbose'])
#     particles_rand = \
#         catalogue_rand._convert_to_cpp_catalogue(verbose=paramset['verbose'])

#     if logger:
#         logger.info(
#             "... prepared catalogue for clustering algorithm.",
#             cpp_state='end'
#         )

#     # Set up constants.
#     alpha = catalogue_data.wstotal / catalogue_rand.wstotal

#     if logger:
#         logger.info("Alpha contrast: %.6e.", alpha)

#     norm_factor_part = _calc_bispec_normalisation_from_particles(
#         particles_rand, alpha
#     )
#     norm_factor_mesh = _calc_bispec_normalisation_from_mesh(
#         particles_rand, paramset, alpha
#     )
#     norm_factor_meshes = 0.

#     if paramset['norm_convention'] == 'none':
#         norm_factor = 1.
#         norm_log_mesg = (
#             "Normalisation factors: "
#             "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; n/a) (none used)."  # noqa: E501
#         )
#     if paramset['norm_convention'] == 'particle':
#         norm_factor = norm_factor_part
#         norm_log_mesg = (
#             "Normalisation factors: "
#             "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed; n/a)."
#         )
#     if paramset['norm_convention'] == 'mesh':
#         norm_factor = norm_factor_mesh
#         norm_log_mesg = (
#             "Normalisation factors: "
#             "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed; n/a)."
#         )
#     if paramset['norm_convention'] == 'mesh-mixed':
#         norm_factor = norm_factor_meshes
#         norm_log_mesg = (
#             "Normalisation factors: "
#             "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; used)."
#         )

#     if logger:
#         logger.info(
#             norm_log_mesg,
#             norm_factor_part, norm_factor_mesh, norm_factor_meshes
#     )

#     # Perform measurement.
#     if logger:
#         logger.info(
#             "Measuring bispectrum from paired survey-type catalogues "
#             "with line-of-sight end-point choice %d "
#             "in the local plane-parallel approximation...",
#             los_choice,
#             cpp_state='start'
#         )

#     results = _compute_bispec_for_los_choice(
#         particles_data, particles_rand, los_data, los_rand, los_choice,
#         paramset, binning, norm_factor
#     )

#     if logger:
#         logger.info(
#             "... measured bispctrum from paired survey-type catalogues "
#             "with line-of-sight end-point choice %d "
#             "in the local plane-parallel approximation.",
#             los_choice,
#             cpp_state='end'
#         )

#     if save:
#         odirpath = paramset['directories']['measurements'] or ""
#         header = "\n".join([
#             catalogue_data.write_attrs_as_header(catalogue_ref=catalogue_rand),  # noqa: E501
#             _print_measurement_header(
#                 paramset,
#                 norm_factor_part, norm_factor_mesh, norm_factor_meshes
#             ),
#         ])
#         if save.lower() == '.txt':
#             datatab = _assemble_measurement_datatab(results, paramset)
#             datafmt = '\t'.join(
#                 (['%.9e'] * 2 + ['%10d']) * 2 +
#                 ['% .9e'] * (datatab.shape[-1] - 6)
#             )
#             ofilename = _get_measurement_filename(paramset)
#             ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
#             np.savetxt(
#                 ofilepath, datatab,
#                 fmt=datafmt, header=header, delimiter='\t'
#             )
#         elif save.lower().endswith('.npz'):
#             results.update({'header': header})
#             ofilename = _get_measurement_filename(paramset)
#             ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
#             np.savez(ofilepath, **results)
#         else:
#             raise ValueError(
#                 f"Unrecognised save format for measurements: {save}."
#             )

#         if logger:
#             logger.info("Measurements saved to %s.", ofilepath)

#     return results


# ========================================================================
# Simulation statistics
# ========================================================================

def _compute_3pt_stats_sim_like(threept_algofunc, catalogue_data,
                                paramset=None, params_sampling=None,
                                degrees=None, binning=None,
                                form=None, idx_bin=None, types=None,
                                save=False, logger=None):
    """Compute three-point statistics from a simulation-box catalogue in
    the global plane-parallel approximation.

    Parameters
    ----------
    threept_algofunc : callable
        Three-point statistic algorithmic function.
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set.  If `None` (default), `degrees`, `binning`,
        `form` and `params_sampling` should be provided; `idx_bin`
        should be provided when `form` is 'full'.
    params_sampling : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    types : dict, optional
        'catalogue_type' and 'statistic_type' values (default is `None`).
        This should be set by the caller of this function.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results.

    """
    # --------------------------------------------------------------------
    # Initialisation
    # --------------------------------------------------------------------

    # -- Parameters ------------------------------------------------------

    paramset = _amalgamate_parameters(
        paramset=paramset, params_sampling=params_sampling,
        degrees=degrees, binning=binning, form=form, idx_bin=idx_bin, **types
    )

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset, logger=logger)

    if logger:
        logger.info("Parameter set have been initialised.")

    if paramset['catalogue_type'] != 'sim':
        raise ValueError(
            "`paramset` 'catalogue_type' does not correspond to "
            "the global plane-parallel clustering algorithm being called: "
            f"catalogue_type = '{paramset['catalogue_type']}'."
        )

    if paramset['statistic_type'] == 'bispec':
        statistic_name = 'bispectrum'
    elif paramset['statistic_type'] == '3pcf':
        statistic_name = 'three-point correlation function'
    else:
        raise ValueError(
            "`paramset` 'statistic_type' does not correspond to "
            "the clustering algorithm being called: "
            f"statistic_type = '{paramset['statistic_type']}'."
        )

    # -- Data ------------------------------------------------------------

    # Set up binning.
    if binning is None:
        binning = Binning.from_parameter_set(paramset)

    if logger:
        logger.info("Binning has been initialised.")

    # Set up box alignment.
    catalogue_data.periodise(
        [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
    )

    if logger:
        logger.info("Catalogue box has been periodised.")

    # --------------------------------------------------------------------
    # Measurements
    # --------------------------------------------------------------------

    # Prepare catalogues.
    if not catalogue_data['nz'].any():
        catalogue_data.compute_mean_density(
            boxsize=list(paramset['boxsize'].values())
        )
        if logger:
            logger.info(
                "Inserted missing 'nz' field "
                "based on particle count and box size."
            )

    if logger:
        logger.info(
            "Preparing catalogue for clustering algorithm...",
            cpp_state='start'
        )

    particles_data = \
        catalogue_data._convert_to_cpp_catalogue(verbose=paramset['verbose'])

    if logger:
        logger.info(
            "... prepared catalogue for clustering algorithm.",
            cpp_state='end'
        )

    # Set up constants.
    norm_factor_part = _calc_bispec_normalisation_from_particles(
        particles_data, alpha=1.
    )
    norm_factor_mesh = _calc_bispec_normalisation_from_mesh(
        particles_data, paramset, alpha=1.
    )
    norm_factor_meshes = 0.

    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; n/a) (none used)."
        )
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed; n/a)."
        )
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed; n/a)."
        )

    if logger:
        logger.info(
            norm_log_mesg,
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        )

    # Perform measurement.
    if logger:
        logger.info(
            "Measuring %s from a simulation-box catalogue "
            "in the global plane-parallel approximation...",
            statistic_name,
            cpp_state='start'
        )

    results = threept_algofunc(particles_data, paramset, binning, norm_factor)

    if logger:
        logger.info(
            "... measured %s from a simulation-box catalogue "
            "in the global plane-parallel approximation.",
            statistic_name,
            cpp_state='end'
        )

    if save:
        odirpath = paramset['directories']['measurements'] or ""
        header = "\n".join([
            catalogue_data.write_attrs_as_header(),
            _print_measurement_header(
                paramset,
                norm_factor_part, norm_factor_mesh, norm_factor_meshes
            ),
        ])
        if save.lower() == '.txt':
            datatab = _assemble_measurement_datatab(results, paramset)
            datafmt = '\t'.join(
                (['%.9e'] * 2 + ['%10d']) * 2 +
                ['% .9e'] * (datatab.shape[-1] - 6)
            )
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
            np.savetxt(
                ofilepath, datatab, fmt=datafmt, header=header, delimiter='\t'
            )
        elif save.lower().endswith('.npz'):
            results.update({'header': header})
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
            np.savez(ofilepath, **results)
        else:
            raise ValueError(
                f"Unrecognised save format for measurements: {save}."
            )

        if logger:
            logger.info("Measurements saved to %s.", ofilepath)

    return results


def compute_bispec_in_gpp_box(catalogue_data,
                              degrees=None, binning=None,
                              form=None, idx_bin=None,
                              sampling_params=None,
                              paramset=None,
                              save=False, logger=None):
    """Compute bispectrum from a simulation-box catalogue in the global
    plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used in lieu of
        `degrees`, `binning`, `form`, `idx_bin` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results as a dictionary with the following entries---

        - 'k1_bin', 'k2_bin': central wavenumber for each bin of
          the first and second wavenumbers;
        - 'k1_eff', 'k2_eff': effective wavenumber for each bin of
          the first and second wavenumbers;
        - 'nmodes_1', 'nmodes_2': number of wavevector modes in each bin
          of the first and second wavenumbers;
        - 'bk_raw': bispectrum raw measurements including any
          specified normalisation and shot noise;
        - 'bk_shot': bispectrum shot noise.

        The effective wavenumber is here defined as the average wavenumber
        in each bin.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.threept.compute_bispec`
    (though without the line-of-sight arguments).

    """
    results = _compute_3pt_stats_sim_like(
        _compute_bispec_in_gpp_box,
        catalogue_data,
        paramset=paramset, params_sampling=sampling_params,
        degrees=degrees, binning=binning, form=form, idx_bin=idx_bin,
        types={'catalogue_type': 'sim', 'statistic_type': 'bispec'},
        save=save, logger=logger
    )

    return results


def compute_3pcf_in_gpp_box(catalogue_data,
                            degrees=None, binning=None,
                            form=None, idx_bin=None,
                            sampling_params=None,
                            paramset=None,
                            save=False, logger=None):
    """Compute three-point correlation function from a simulation-box
    catalogue in the global plane-parallel approximation.

    Parameters
    ----------
    catalogue_data : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Data-source catalogue.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used in lieu of
        `degree`, `binning`, `form`, `idx_bin` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results as a dictionary with the following entries---

        - 'r1_bin', 'r2_bin': central separation for each bin of
          the first and second separations;
        - 'r1_eff', 'r2_eff': effective separation for each bin of
          the first and second separations;
        - 'npairs_1', 'npairs_2': number of separation pairs in each bin
          of the first and second separations;
        - 'zeta_raw': three-point correlation function raw measurements
          including any specified normalisation and shot noise;
        - 'zeta_shot': three-point correlation function shot noise.

        The effective separation is here defined as the average separation
        in each bin.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.threept.compute_bispec`
    (though without the line-of-sight arguments).

    """
    results = _compute_3pt_stats_sim_like(
        _compute_3pcf_in_gpp_box,
        catalogue_data,
        paramset=paramset, params_sampling=sampling_params,
        degrees=degrees, binning=binning, form=form, idx_bin=idx_bin,
        types={'catalogue_type': 'sim', 'statistic_type': '3pcf'},
        save=save, logger=logger
    )

    return results


# ========================================================================
# Window statistics
# ========================================================================

def compute_3pcf_window(catalogue_rand, los_rand=None,
                        degrees=None, wa_orders=None,
                        binning=None, form=None, idx_bin=None,
                        sampling_params=None,
                        paramset=None,
                        save=False, logger=None):
    """Compute three-point correlation function window from a
    random catalogue.

    Parameters
    ----------
    catalogue_rand : :class:`~triumvirate.catalogue.ParticleCatalogue`
        Random-source catalogue.
    los_rand : (N, 3) array of float, optional
        Specified lines of sight for the random-source catalogue.
        If `None` (default), this is automatically computed using
        :meth:`~triumvirate.catalogue.ParticleCatalogue.compute_los`.
    degrees : tuple of (int, int, int) or str of length 3, optional
        Multipole degrees either as a tuple ('ell1', 'ell2', 'ELL') or
        as a string of length 3.  If not `None` (default), this will
        override ``paramset['degrees']`` entries.  If a string, multipole
        degrees are assumed to be single-digit integers.
    wa_orders : tuple of (int, int) or str of length 2, optional
        Wide-angle correction orders either as a tuple ('i_wa', 'j_wa') or
        as a string of length 2.  If not `None` (default), this will
        override ``paramset['wa_orders']`` entries.  If a string,
        multipole degrees are assumed to be single-digit integers.
    binning : :class:`~triumvirate.dataobjs.Binning`, optional
        Binning for the measurements.  If `None` (default),
        this is constructed from `paramset`.
    form : {'diag', 'off-diag', 'row', 'full'}, optional
        Binning form of the measurements.  If not `None` (default),
        this will override ``paramset['form']``.

        .. versionchanged:: 0.3.0
            Add 'off-diag' and 'row' options and redefine 'full' option.

    idx_bin : int, optional
        When binning `form` is 'row', this is the fixed bin index for
        the first coordinate dimension; when binning `form` is 'off-diag',
        this is the upper-triangular off-diagonal index; otherwise, this
        is ignored.  If not `None` (default), this will override
        ``paramset['idx_bin']``.

        .. versionchanged:: 0.3.0
            Expand definition to match changed `form`.

    sampling_params : dict, optional
        Dictionary containing a subset of the following entries
        for sampling parameters---

        - 'alignment': {'centre', 'pad'};
        - 'boxsize': sequence of [float, float, float];
        - 'ngrid': sequence of [int, int, int];
        - 'assignment': {'ngp', 'cic', 'tsc', 'pcs'};
        - 'interlace': bool;

        and exactly one of the following only when 'alignment' is 'pad'---

        - 'boxpad': float;
        - 'gridpad': float.

        If not `None` (default), this will override the corresponding
        entries in `paramset`.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Full parameter set (default is `None`).  This is used in lieu of
        `degrees`, `binning`, `form`, `idx_bin` or `sampling_params`.
    save : {'.txt', '.npz', False}, optional
        If not `False` (default), save the measurements as a '.txt' file
        or in '.npz' format.  The save path is determined from `paramset`
        (if unset, a default file path in the current working directory is
        used).
    logger : :class:`logging.Logger`, optional
        Logger (default is `None`).

    Returns
    -------
    results : dict of {str: :class:`numpy.ndarray`}
        Measurement results as a dictionary with the following entries---

        - 'r1_bin', 'r2_bin': central separation for each bin of
          the first and second separations;
        - 'r1_eff', 'r2_eff': effective separation for each bin of
          the first and second separations;
        - 'npairs_1', 'npairs_2': number of separation pairs in each bin
          of the first and second separations;
        - 'zeta_raw': three-point correlation function raw measurements
          including any specified normalisation and shot noise;
        - 'zeta_shot': three-point correlation function shot noise.

        The effective separation is here defined as the average separation
        in each bin.

    Examples
    --------
    See analogous examples in :func:`~triumvirate.threept.compute_bispec`.

    """
    # --------------------------------------------------------------------
    # Initialisation
    # --------------------------------------------------------------------

    # -- Parameters ------------------------------------------------------

    if wa_orders is not None:
        wide_angle = True
    else:
        wide_angle = False

    paramset = _amalgamate_parameters(
        paramset=paramset, params_sampling=sampling_params,
        degrees=degrees, wa_orders=wa_orders,
        binning=binning, form=form, idx_bin=idx_bin,
        catalogue_type='random', statistic_type='3pcf-win'
    )

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset, logger=logger)

    if logger:
        logger.info("Parameter set have been initialised.")

    # -- Data ------------------------------------------------------------

    # Set up binning.
    if binning is None:
        binning = Binning.from_parameter_set(paramset)

    if logger:
        logger.info("Binning has been initialised.")

    # Set up lines of sight.
    if los_rand is None:
        los_rand = catalogue_rand.compute_los()
    los_rand = np.ascontiguousarray(los_rand)

    if logger:
        logger.info("Lines of sight have been initialised.")

    # Set up box alignment.
    catalogue_rand.centre(
        [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
    )

    catalogue_rand.periodise(
        [paramset['boxsize'][ax] for ax in ['x', 'y', 'z']]
    )

    if logger:
        logger.info("Catalogues have been aligned.")

    # --------------------------------------------------------------------
    # Measurements
    # --------------------------------------------------------------------

    # Prepare catalogues.
    particles_rand = \
        catalogue_rand._convert_to_cpp_catalogue(verbose=paramset['verbose'])

    # Set up constants.
    norm_factor_part = _calc_bispec_normalisation_from_particles(
        particles_rand, alpha=1.
    )
    norm_factor_mesh = _calc_bispec_normalisation_from_mesh(
        particles_rand, paramset, alpha=1.
    )
    norm_factor_meshes = 0.

    if paramset['norm_convention'] == 'none':
        norm_factor = 1.
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh), %.6e (mesh-mixed; n/a) (none used)."
        )
    if paramset['norm_convention'] == 'particle':
        norm_factor = norm_factor_part
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle; used), %.6e (mesh), %.6e (mesh-mixed; n/a)."
        )
    if paramset['norm_convention'] == 'mesh':
        norm_factor = norm_factor_mesh
        norm_log_mesg = (
            "Normalisation factors: "
            "%.6e (particle), %.6e (mesh; used), %.6e (mesh-mixed; n/a)."
        )

    if logger:
        logger.info(
            norm_log_mesg,
            norm_factor_part, norm_factor_mesh, norm_factor_meshes
        )

    # Perform measurement.
    if logger:
        logger.info(
            "Measuring three-point correlation function window "
            "from a random catalogue...",
            cpp_state='start'
        )

    results = _compute_3pcf_window(
        particles_rand, los_rand,
        paramset, binning, alpha=1., norm_factor=norm_factor,
        wide_angle=wide_angle
    )

    if logger:
        logger.info(
            "... measured three-point correlation function window "
            "from a random catalogue.",
            cpp_state='end'
        )

    if save:
        odirpath = paramset['directories']['measurements'] or ""
        header = "\n".join([
            catalogue_rand.write_attrs_as_header(),
            _print_measurement_header(
                paramset,
                norm_factor_part, norm_factor_mesh, norm_factor_meshes
            ),
        ])
        if save.lower() == '.txt':
            datatab = _assemble_measurement_datatab(results, paramset)
            datafmt = '\t'.join(
                (['%.9e'] * 2 + ['%10d']) * 2 +
                ['% .9e'] * (datatab.shape[-1] - 6)
            )
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.txt')
            np.savetxt(
                ofilepath, datatab, fmt=datafmt, header=header, delimiter='\t'
            )
        elif save.lower().endswith('.npz'):
            results.update({'header': header})
            ofilename = _get_measurement_filename(paramset)
            ofilepath = Path(odirpath, ofilename).with_suffix('.npz')
            np.savez(ofilepath, **results)
        else:
            raise ValueError(
                f"Unrecognised save format for measurements: {save}."
            )

        if logger:
            logger.info("Measurements saved to %s.", ofilepath)

    return results



"""Test :mod:`~triumvirate.fieldmesh`.

"""
import numpy as np
import pytest

from triumvirate.fieldmesh import record_binned_vectors


@pytest.mark.parametrize(
    "case, statistic_type, count, last_entry",
    [
        (
            'default', 'modes', 19548,
            (3, 0.08, 0.105, 0.10053096, 0.02513274, 0.01256637),
        ),
        (
            'override', 'modes', 2468,
            (3, 0.08, 0.105, 0.10053096, 0.02513274, 0.01256637),
        ),
        (
            'empty', 'pairs', 228396,
            (3, 125., 150., 148.4375, 19.53125, 7.8125),
        ),
    ]
)
def test_record_binned_vectors(case, statistic_type, count, last_entry,
                               request):

    if case == 'default':
        test_paramset = request.getfixturevalue('test_paramset')
        test_paramset.update(statistic_type=statistic_type)
        boxsize = None
        ngrid = None
    elif case == 'override':
        test_paramset = request.getfixturevalue('test_paramset')
        test_paramset.update(statistic_type=statistic_type)
        boxsize = 500.
        ngrid = 128
    elif case == 'empty':
        test_paramset = None
        boxsize = 500.
        ngrid = 128
    else:
        raise ValueError("Invalid test case for `record_binned_vectors`.")

    if statistic_type == 'modes':
        test_binning = request.getfixturevalue('test_binning_fourier')
    elif statistic_type == 'pairs':
        test_binning = request.getfixturevalue('test_binning_config')
    else:
        raise ValueError("Invalid statistic type for `record_binned_vectors`.")

    binned_vectors = record_binned_vectors(
        test_binning, paramset=test_paramset, boxsize=boxsize, ngrid=ngrid
    )
    sorting = np.lexsort((
        binned_vectors['vecz'],
        binned_vectors['vecy'],
        binned_vectors['vecx'],
        binned_vectors['index'],
    ))

    binned_vectors_sorted = binned_vectors[sorting]

    assert len(binned_vectors_sorted) == count, (
        "Binned vectors have incorrect count."
    )
    for ifield, name in enumerate(binned_vectors_sorted.dtype.names):
        assert np.allclose(
            binned_vectors_sorted[-1][ifield], last_entry[ifield]
        ), (
            "The last entry of binned vectors has "
            f"incorrect '{name}' field value."
        )



"""
Hankel-Like Transforms (:mod:`~triumvirate.transforms`)
=======================================================

.. versionadded:: 0.4.6


Perform Hankel-like transforms in 1- and 2-d.

.. autosummary::
    SphericalBesselTransform
    DoubleSphericalBesselTransform
    resample_lglin
    resample_lin

"""  # numpydoc ignore=SS01
from collections.abc import Sequence

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline, RectBivariateSpline

from ._arrayops import (
    extrap_lin,
    extrap_loglin,
    extrap_loglin_oscil,
    extrap_pad,
    extrap2d_lin,
    extrap2d_loglin,
    extrap2d_loglin_oscil,
    extrap2d_pad,
)
from ._fftlog import HankelTransform


class SphericalBesselTransform:
    """Spherical Bessel Transform.

    Parameters
    ----------
    degree : int
        Degree of the spherical Bessel transform.
    bias : int
        Power-law bias index.
    sample_pts : array of float
        Pre-transform sample points in 1-d.  Must be
        logarithmically spaced.  Must be even in length if extrapolation
        is used.
    pivot : float, optional
        Pivot value (default is 1.).  When `lowring` is `True`, this is
        adjusted if it is non-zero, or otherwise directly calculated.
    lowring : bool, optional
        Low-ringing condition (default is `True`).
    extrap : int, optional
        Extrapolation method (default is 0) with the following
        options:

        - 0: none;
        - 1: extrapolate by zero padding;
        - 2: extrapolate by constant padding;
        - 3: extrapolate linearly;
        - 4: extrapolate log-linearly;
        - 5: extrapolate log-linearly with oscillatory behaviour.

        Any extrapolation results in a sample size for the transform that
        is the smallest power of 2 greater than or equal to `extrap_exp`
        times the original number of sample points; the pre-transform
        samples are assumed to be real.
    extrap_exp : float, optional
        Sample size expansion factor (default is 2.) for extrapolation.
        The smallest power of 2 greater than or equal to this times the
        original number of sample points is used as the sample size for
        the transform.
    extrap_layer : {'native', 'outer'}, optional
        Extrapolation layer:

        - 'native': extrapolate the samples, including
          any pre-factors, within the underlying FFTLog algorithm;
        - 'outer' (default): extrapolate the samples, excluding
          any pre-factors, before passing them to the underlying
          FFTLog algorithm.

    threaded : bool, optional
        If `True` (default is `False`), use the multi-threaded FFTLog
        algorithm.

    Attributes
    ----------
    degree : int
        Degree of the spherical Bessel transform.
    bias : int
        Power-law bias index.
    size : int
        Sample size.
    pivot : float
        Pivot value used.

    Raises
    ------
    ValueError
        When `extrap` is not a supported option.
    ValueError
        When the input sample size is not even and extrapolation is used.

    """

    def __init__(self, degree, bias, sample_pts, pivot=1., lowring=True,
                 extrap=0, extrap_exp=2., extrap_layer='outer',
                 threaded=False):
        if extrap not in {0, 1, 2, 3, 4, 5}:
            raise ValueError(
                f"Extrapolation option must be in 0 to 5: {extrap=}."
            )

        self.degree = degree
        self.bias = bias

        self._lowring = lowring
        self._extrap = extrap
        self._extrap_layer = extrap_layer

        if self._extrap:
            if (nsamp := len(sample_pts)) % 2:
                raise ValueError(
                    "Input sample size must be even when extrapolation "
                    f"is used: {len(sample_pts)=}."
                )
            if self._extrap_layer == 'native':
                extrap_native = self._extrap
                self._n_ext = 0
            elif self._extrap_layer == 'outer':
                extrap_native = 0
                nsamp_trans = int(2 ** np.ceil(np.log2(extrap_exp * nsamp)))
                self._n_ext = (nsamp_trans - nsamp) // 2
                sample_pts = extrap_loglin(sample_pts, self._n_ext)
            else:
                raise ValueError(
                    "Extrapolation layer must be 'native' or 'outer': "
                    f"extrap_layer={self._extrap_layer}."
                )
        else:
            self._n_ext = 0
            extrap_native = 0

        self._fbht = HankelTransform(
            degree + 1./2, bias, sample_pts,
            kr_c=pivot, lowring=lowring,
            extrap=extrap_native, extrap_exp=extrap_exp,
            threaded=threaded
        )

        self._post_slice = slice(self._n_ext, -self._n_ext) if self._n_ext \
            else slice(None)

        self._logres = self._fbht._logres
        self._pre_sampts = np.asarray(self._fbht._pre_sampts)
        self._post_sampts = np.asarray(self._fbht._post_sampts)[
            self._post_slice
        ]

    @property
    def size(self):
        """Sample size of the transform.

        Returns
        -------
        int

        """  # numpydoc ignore=RT03
        return self._fbht.size

    @property
    def pivot(self):
        """Pivot value.

        Returns
        -------
        float

        """  # numpydoc ignore=RT03
        return self._fbht._pivot

    def transform(self, pre_samples):
        """Transform samples at initialised sample points.

        Parameters
        ----------
        pre_samples : array_like
            Pre-transform samples in 1-d.  Assumed to be real
            if extrapolation is used.

        Returns
        -------
        post_sampts, post_samples : array_like
            Post-transform sample points and samples.

        """
        if self._extrap_layer == 'outer':
            pre_samples = self._perform_extrap(pre_samples)

        pre_samples *= self._pre_sampts ** (3./2)

        post_sampts, post_samples = self._fbht.transform(pre_samples)

        post_sampts = post_sampts[self._post_slice]
        post_samples = post_samples[self._post_slice]

        post_samples *= (2*np.pi / self._post_sampts) ** (3./2)

        return post_sampts, post_samples

    def transform_cosmo_multipoles(self, direction, pre_samples):
        """Transform cosmological multipoles between configuration and
        Fourier spaces.

        Parameters
        ----------
        direction : {'forward', 'backward', 1, -1}
            Transform direction: 'forward' or 1 from configuration to
            Fourier space, 'backward' or -1 from Fourier to configuration
            space.
        pre_samples : array of float
            Pre-transform samples in 1-d.  Assumed to be real
            if extrapolation is used.

        Returns
        -------
        post_sampts, post_samples : array of float
            Post-transform sample points and samples.

        """
        if direction in {'forward', 1}:
            parity = (-1j) ** self.degree
            prefactor = 1.
        elif direction in {'backward', -1}:
            parity = (1j) ** self.degree
            prefactor = (2*np.pi) ** (-3)
        else:
            raise ValueError(
                "Transform direction must be 'forward'/1 or 'backward'/-1."
            )

        post_sampts, post_samples = self.transform(pre_samples)

        post_samples *= prefactor * parity

        return post_sampts, post_samples

    def _perform_extrap(self, arr):
        """Perform extrapolation according to internal attributes.

        Parameters
        ----------
        arr : array_like
            1-d array to extrapolate.

        Returns
        -------
        array_like
            Extrapolated 1-d array.

        """
        if self._extrap == 1:
            return extrap_pad(arr, self._n_ext, 0., 0.)
        if self._extrap == 2:
            return extrap_pad(arr, self._n_ext, arr[0], arr[-1])
        if self._extrap == 3:
            return extrap_lin(arr, self._n_ext)
        if self._extrap == 4:
            return extrap_loglin(arr, self._n_ext)
        if self._extrap == 5:
            return extrap_loglin_oscil(arr, self._n_ext)
        return arr


class DoubleSphericalBesselTransform:
    """Double spherical Bessel Transform.

    Parameters
    ----------
    degrees : (int, int)
        Degrees of the double spherical Bessel transform.
    biases : (int, int)
        Power-law bias indices.
    sample_pts : array of float
        Pre-transform sample points in 1-d for both dimensions.
        Must be logarithmically spaced.  Must be even in length if
        extrapolation is used.
    pivot : float, optional
        Pivot value (default is 1.).  When `lowring` is `True`, this is
        adjusted if it is non-zero, or otherwise directly calculated.
    lowring : bool, optional
        Low-ringing condition (default is `True`).
    extrap : int or tuple of int, optional
        Extrapolation method for both or each dimension(s) (default is 0)
        with the following options:

        - 0: none;
        - 1: extrapolate by zero padding;
        - 2: extrapolate by constant padding;
        - 3: extrapolate linearly;
        - 4: extrapolate log-linearly;
        - 5: extrapolate log-linearly with oscillatory behaviour.

        Any extrapolation results in a sample size for the transform that
        is the smallest power of 2 greater than or equal to `extrap_exp`
        times the original number of sample points; the pre-transform
        samples are assumed to be real.
    extrap_exp : float, optional
        Sample size expansion factor (default is 2.) for extrapolation.
        The smallest power of 2 greater than or equal to this times the
        original number of sample points is used as the sample size for
        the transform.
    extrap2d : bool, optional
        If `True` (default is `False`), perform 2-d extrapolation
        pre-transform excluding any pre-factors; otherwise, perform 1-d
        extrapolation including any pre-factors within the underlying
        FFTLog algorithm.
    threaded : bool, optional
        If `True` (default is `False`), use the multi-threaded FFTLog
        algorithm.

    Attributes
    ----------
    degrees : (int, int)
        Degrees of the spherical Bessel transform.
    biases : (int, int)
        Power-law bias indices.
    size : int
        Sample size for either dimension.
    pivot : (float, float)
        Pivot values used for both dimensions.

    Raises
    ------
    ValueError
        When `extrap` is not a supported option.
    ValueError
        When the input sample size is not even and extrapolation is used.

    """
    # Representative transform, either vertical (0, across rows)
    # or horizontal (1, across columns).
    _rep = 0

    def __init__(self, degrees, biases, sample_pts, pivot=1., lowring=True,
                 extrap=0, extrap_exp=2., extrap2d=False, threaded=False):
        if extrap not in {0, 1, 2, 3, 4, 5}:
            raise ValueError(
                f"Extrapolation option must be in 0 to 5: {extrap=}."
            )

        self.degrees = degrees
        self.biases = biases

        self._lowring = lowring
        self._extrap = extrap
        self._extrap2d = extrap2d

        if self._extrap:
            if (nsamp := len(sample_pts)) % 2:
                raise ValueError(
                    "Input sample size must be even when extrapolation "
                    f"is used: {len(sample_pts)=}."
                )
            if not self._extrap2d:
                extrap_native = self._extrap
                self._n_ext = 0
            else:
                extrap_native = 0
                nsamp_trans = int(2 ** np.ceil(np.log2(extrap_exp * nsamp)))
                self._n_ext = (nsamp_trans - nsamp) // 2
                sample_pts = extrap_loglin(sample_pts, self._n_ext)
        else:
            self._n_ext = 0
            extrap_native = 0

        self._fbsjt = tuple(
            SphericalBesselTransform(
                degree_, bias_, sample_pts,
                pivot=pivot, lowring=lowring,
                extrap=extrap_native, extrap_exp=extrap_exp,
                threaded=threaded
            )
            for (degree_, bias_) in zip(degrees, biases)
        )

        self._post_slice = slice(self._n_ext, -self._n_ext) if self._n_ext \
            else slice(None)

        self._logres = self._fbsjt[self._rep]._logres
        self._pre_sampts = np.asarray(self._fbsjt[self._rep]._pre_sampts)
        self._post_sampts = np.asarray(self._fbsjt[self._rep]._post_sampts)[
            self._post_slice
        ]

    @property
    def size(self):
        """Sample size of the transform for either dimension.

        Returns
        -------
        int

        """  # numpydoc ignore=RT03
        return self._fbsjt[self._rep].size

    @property
    def pivot(self):
        """Pivot values for both dimensions.

        Returns
        -------
        tuple of float

        """  # numpydoc ignore=RT03
        return (self._fbsjt[0]._pivot, self._fbsjt[1]._pivot)

    def transform(self, pre_samples):
        """Transform samples at initialised sample points.

        Parameters
        ----------
        pre_samples : array_like
            Pre-transform samples in 2-d.  Assumed to be real
            if extrapolation is used.

        Returns
        -------
        post_sampts, post_samples : array_like
            Post-transform sample points and samples in 2-d, with
            `post_sampts` in :func:`numpy.meshgrid` 'ij'-indexing format.

        """
        if self._extrap2d:
            pre_samples = self._perform_extrap(pre_samples)

        inter_samples = []
        for pre_samples_row in pre_samples:
            _, inter_samples_row = self._fbsjt[1].transform(pre_samples_row)
            inter_samples.append(inter_samples_row)

        post_samples_T = []
        for inter_samples_col in np.asarray(inter_samples).T:
            _, post_samples_col = self._fbsjt[0].transform(inter_samples_col)
            post_samples_T.append(post_samples_col)

        post_samples = np.asarray(post_samples_T).T
        post_samples = post_samples[self._post_slice, self._post_slice]

        post_sampts = np.meshgrid(
            self._post_sampts, self._post_sampts, indexing='ij'
        )

        return post_sampts, post_samples

    def transform_cosmo_multipoles(self, direction, pre_samples):
        """Transform cosmological multipoles between configuration and
        Fourier spaces.

        Parameters
        ----------
        direction : {'forward', 'backward', 1, -1}
            Transform direction: 'forward' or 1 from configuration to
            Fourier space, 'backward' or -1 from Fourier to configuration
            space.
        pre_samples : array of float
            Pre-transform samples in 2-d.  Assumed to be real
            if extrapolation is used.

        Returns
        -------
        post_sampts, post_samples : array of float
            Post-transform sample points and samples in 2-d, with
            `post_sampts` in :func:`numpy.meshgrid` 'ij'-indexing format.

        """
        if direction in {'forward', 1}:
            parity = (-1j) ** sum(self.degrees)
            prefactor = 1.
        elif direction in {'backward', -1}:
            parity = (1j) ** sum(self.degrees)
            prefactor = (2*np.pi) ** (-3 * len(self.degrees))
        else:
            raise ValueError(
                "Transform direction must be 'forward'/1 or 'backward'/-1."
            )

        post_sampts, post_samples = self.transform(pre_samples)

        post_samples *= prefactor * parity

        return post_sampts, post_samples

    def _perform_extrap(self, arr):
        """Perform extrapolation according to internal attributes.

        Parameters
        ----------
        arr : array_like
            2-d array to extrapolate.

        Returns
        -------
        array_like
            Extrapolated 2-d array.

        """
        arr = np.asarray(arr)
        if self._extrap == 1:
            return extrap2d_pad(
                arr, self._n_ext, 0., 0., 0., 0.
            )
        if self._extrap == 2:
            return extrap2d_pad(
                arr, self._n_ext, arr[:, 0], arr[:, -1], arr[0, :], arr[-1, :]
            )
        if self._extrap == 3:
            return extrap2d_lin(arr, self._n_ext)
        if self._extrap == 4:
            return extrap2d_loglin(arr, self._n_ext)
        if self._extrap == 5:
            return extrap2d_loglin_oscil(arr, self._n_ext)
        return arr


def resample_lglin(sampts, samples, size=None, spline=3):
    """Resample at logarithmically spaced sample points in 1- or 2-d.

    Parameters
    ----------
    sampts : array_like or sequence of array_like
        Original sample points (1-d in each dimension).
    samples : array_like
        Original samples in 1-d or 2-d.
    size : int or sequence of int, optional
        Size of the interpolated samples.  If `None` (default), the size
        is kept the same as the original samples.
    spline : int, optional
        Degree of the interpolation spline (default is 3).

    Returns
    -------
    resampts : (2-tuple of) :class:`numpy.ndarray`
        Logarithmically spaced sample points (1-d for each dimension)
        in the original range(s).
    resamples : :class:`numpy.ndarray`
        Interpolated samples.

    Examples
    --------
    Resample a 2-d function:

    >>> def f(x, y): return np.exp(-x*np.sqrt(y))
    >>> sampts = 1.e-3 * np.arange(1., 51.)
    >>> samples = f(*np.meshgrid(sampts, sampts, indexing='ij'))
    >>> resampts, resamples = resample_lglin(sampts, samples, size=10)

    Check that the resampling is accurate:

    >>> resamples_expected = f(
    ...     *np.meshgrid(resampts[0], resampts[-1], indexing='ij')
    ... )
    >>> print(
    ...     "Resampling is accurate:",
    ...     np.allclose(resamples, resamples_expected)
    ... )
    Resampling is accurate: True

    """
    sampts = np.squeeze(sampts)
    samples = np.squeeze(samples)

    if samples.ndim == 1:
        if sampts.ndim != 1:
            raise ValueError("Sample points must be 1-d.")

        resampts = np.logspace(
            np.log10(sampts[0]), np.log10(sampts[-1]), size or len(sampts),
            base=10
        )
        resamples = InterpolatedUnivariateSpline(
            sampts, samples, k=spline, ext='raise'
        )(resampts)

        return resampts, resamples

    if samples.ndim == 2:
        if sampts.ndim == 1:
            sampts = np.array([sampts, sampts])
        if len(sampts) != 2 or sampts[0].ndim != 1 or sampts[-1].ndim != 1:
            raise ValueError(
                "Must provide (two) 1-d array(s) as sample points "
                "for 2-d resampling."
            )

        size = size if isinstance(size, Sequence) else (size, size)

        resampts = (
            np.logspace(
                np.log10(sampts[0][0]), np.log10(sampts[0][-1]),
                size[0] or len(sampts[0]), base=10
            ),
            np.logspace(
                np.log10(sampts[-1][0]), np.log10(sampts[-1][-1]),
                size[-1] or len(sampts[-1]), base=10
            )
        )
        resamples = RectBivariateSpline(
            sampts[0], sampts[-1], samples, kx=spline, ky=spline
        )(*resampts)

        return resampts, resamples

    raise ValueError(f"Sample points must be 1-d or 2-d: {samples.ndim=}.")


def resample_lin(sampts, samples, size=None, spline=3):
    """Resample at linearly spaced sample points in 1- or 2-d.

    Parameters
    ----------
    sampts : array_like or sequence of array_like
        Original sample points (1-d in each dimension).
    samples : array_like
        Original samples in 1-d or 2-d.
    size : int or sequence of int, optional
        Size of the interpolated samples.  If `None` (default), the size
        is kept the same as the original samples.
    spline : int, optional
        Degree of the interpolation spline (default is 3).

    Returns
    -------
    resampts : (2-tuple of) :class:`numpy.ndarray`
        Linearly spaced sample points (1-d for each dimension)
        in the original range(s).
    resamples : :class:`numpy.ndarray`
        Interpolated samples.

    Examples
    --------
    Resample a 2-d function:

    >>> def f(x, y): return np.exp(-x*np.sqrt(y))
    >>> sampts = 1.e-3 * np.arange(1., 51.)
    >>> samples = f(*np.meshgrid(sampts, sampts, indexing='ij'))
    >>> resampts, resamples = resample_lin(sampts, samples, size=10)

    Check that the resampling is accurate:

    >>> resamples_expected = f(
    ...     *np.meshgrid(resampts[0], resampts[-1], indexing='ij')
    ... )
    >>> print(
    ...     "Resampling is accurate:",
    ...     np.allclose(resamples, resamples_expected)
    ... )
    Resampling is accurate: True

    """
    sampts = np.squeeze(sampts)
    samples = np.squeeze(samples)

    if samples.ndim == 1:
        if sampts.ndim != 1:
            raise ValueError("Sample points must be 1-d.")

        resampts = np.linspace(
            sampts[0], sampts[-1], size or len(sampts)
        )
        resamples = InterpolatedUnivariateSpline(
            sampts, samples, k=spline, ext='raise'
        )(resampts)

        return resampts, resamples

    if samples.ndim == 2:
        if sampts.ndim == 1:
            sampts = np.array([sampts, sampts])
        if len(sampts) != 2 or sampts[0].ndim != 1 or sampts[-1].ndim != 1:
            raise ValueError(
                "Must provide (two) 1-d array(s) as sample points "
                "for 2-d resampling."
            )

        size = size if isinstance(size, Sequence) else (size, size)

        resampts = (
            np.linspace(
                sampts[0][0], sampts[0][-1],
                size[0] or len(sampts[0])
            ),
            np.linspace(
                sampts[-1][0], sampts[-1][-1],
                size[-1] or len(sampts[-1])
            )
        )
        resamples = RectBivariateSpline(
            sampts[0], sampts[-1], samples, kx=spline, ky=spline
        )(*resampts)

        return resampts, resamples

    raise ValueError(f"Sample points must be 1-d or 2-d: {samples.ndim=}.")



"""Test :mod:`~triumvirate.parameters`.

"""
import os.path
import warnings
from collections.abc import Collection, Sequence
from copy import deepcopy
from pprint import pformat

import pytest
import yaml

from triumvirate.parameters import (
    InvalidParameterError,
    ParameterSet,
    fetch_paramset_template,
    _TMPL_PARAM_DICT,
)


def make_template_parameters_valid(tmpl_params):
    """Construct valid parameters from template parameters.

    Parameters
    ----------
    tmpl_params : str or dict
        Template parameters.

    Returns
    -------
    str or dict
        Valid parameters constructed from the template parameters.

    Raises
    ------
    TypeError
        If `template_parameters` type is not :class:`str`
        or :class:`dict`.

    """
    if isinstance(tmpl_params, str):
        def replace_parameter(_text, str_original, str_new):
            return _text.replace(str_original, str_new)

        param_text = tmpl_params
        parameter_replacements = [
            (
                "boxsize:\n  x:\n  y:\n  z:\n",
                "boxsize:\n  x: 1000.\n  y: 1000.\n  z: 1000.\n"
            ),
            (
                "ngrid:\n  x:\n  y:\n  z:\n",
                "ngrid:\n  x: 64\n  y: 64\n  z: 64\n"
            ),
            (
                "catalogue_type:",
                "catalogue_type: sim"
            ),
            (
                "statistic_type:",
                "statistic_type: bispec"
            ),
            (
                "ell1:",
                "ell1: 0"
            ),
            (
                "ell2:",
                "ell2: 0"
            ),
            (
                "ELL:",
                "ELL: 0"
            ),
            (
                "range: [~, ~]",
                "range: [0.005, 0.105]"
            ),
            (
                "num_bins:",
                "num_bins: 10"
            ),
            # Not necessary, but suppresses logging.
            (
                "verbose: 20",
                "verbose: 60"
            ),
        ]

        for _replacement in parameter_replacements:
            param_text = replace_parameter(param_text, *_replacement)

        return param_text

    if isinstance(tmpl_params, dict):
        param_dict = deepcopy(tmpl_params)

        for ax_name in ['x', 'y', 'z']:
            param_dict['boxsize'][ax_name] = 1000.
            param_dict['ngrid'][ax_name] = 64.

        param_dict.update({
            'catalogue_type': 'sim',
            'statistic_type': 'bispec',
            'degrees': {'ell1': 0, 'ell2': 0, 'ELL': 0},
            'range': [0.005, 0.105],
            'num_bins': 10,
            'verbose': 60,  # not necessary but suppresses logging
        })

        return param_dict

    raise TypeError("`template_parameters` is not str or dict.")


# Returns the subset of default parameters from the template.
@pytest.fixture
def default_parameters():
    return {
        'alignment': 'centre',
        'padscale': 'box',
        'assignment': 'tsc',
        'interlace': False,
        'form': 'diag',
        'norm_convention': 'particle',
        'binning': 'lin',
        'fftw_scheme': 'measure',
        'use_fftw_wisdom': False,
        'save_binned_vectors': False,
        'verbose': 20,
    }


@pytest.fixture
def template_parameters_text():
    return fetch_paramset_template('text')


@pytest.fixture
def template_parameters_dict():
    return fetch_paramset_template('dict')


@pytest.mark.parametrize(
    "source, ret_defaults",
    [
        ('text', False),
        ('dict', True),
    ]
)
def test_fetch_paramset_template(source, ret_defaults, default_parameters):

    fetched = fetch_paramset_template(source, ret_defaults=ret_defaults)
    if ret_defaults:
        template, defaults = fetched
    else:
        template = fetched

    # Check for 'text' source and exit early; else assume 'dict' source.
    if source == 'text':
        return
    if source != 'dict':
        raise ValueError("Invalid `pytest.mark.parametrize` parameters.")

    # Check `template` agrees with an independent, private template
    # parameter dictionary.
    assert template == _TMPL_PARAM_DICT, \
        "Fetched template parameters do not match internal records."

    # Check `defaults`` are pre-set default parameters.
    assert defaults == default_parameters, (
        "Fetched template default parameters do not match "
        "pre-set default parameters."
    )

    # Check `template` minus `defaults` are unset.
    non_defaults_keys = set(template.keys()) - set(defaults.keys())
    for param_name in non_defaults_keys:
        param_val = template[param_name]
        if isinstance(param_val, Sequence):
            assert all(
                param_val_element is None for param_val_element in param_val
            ), (
                "Non-NoneType element found in "
                "sequence-type non-default parameter value "
                "in parameter template."
            )
        elif isinstance(param_val, Collection):
            assert all(
                param_val_val is None for param_val_val in param_val.values()
            ), (
                "Non-NoneType value found in "
                "collection-type non-default parameter value "
                "in parameter template."
            )
        else:
            assert param_val is None, (
                "Non-NoneType non-default parameter value "
                "found in parameter template."
            )


@pytest.mark.parametrize(
    "param_filepath, param_dict",
    [
        ("tmpl_params.yml", None),
        (None, 'template_parameters_dict'),
        (
            "tmpl_params.yml",
            'template_parameters_dict',
        ),
    ]
)
def test_ParameterSet___cinit__(param_filepath, param_dict, test_param_dir,
                                request, tmp_path):

    # Patch paths for wheel testing.
    if param_filepath is not None:
        param_filepath = os.path.join(test_param_dir, param_filepath)

    # Convert parametrized parameters to fixtures.
    if param_dict is not None:
        param_dict = request.getfixturevalue(param_dict)

    # Check initialisation fails (and exit early if needed).
    if param_filepath is not None and param_dict is not None:
        with pytest.raises(ValueError):
            ParameterSet(
                param_filepath=param_filepath, param_dict=param_dict
            )
        return

    with pytest.raises(InvalidParameterError):
        ParameterSet(
            param_filepath=param_filepath, param_dict=param_dict
        )

    # Modify failed initialisation and check it succeeds.
    if param_filepath is None and param_dict is not None:
        param_dict = make_template_parameters_valid(param_dict)
        ParameterSet(param_dict=param_dict)
    elif param_filepath is not None and param_dict is None:
        with open(param_filepath) as tmpl_parameter_file:
            param_text = make_template_parameters_valid(
                tmpl_parameter_file.read()
            )

        param_filepath = tmp_path/"params.yml"
        with open(param_filepath, 'w') as param_file:
            param_file.write(param_text)

        ParameterSet(param_filepath=param_filepath)
    else:
        raise ValueError("Invalid `pytest.mark.parametrize` parameters.")


@pytest.mark.parametrize(
    "template_parameter_source",
    [
        'template_parameters_text',
        'template_parameters_dict',
    ]
)
def test_ParameterSet___str__(template_parameter_source, request, tmp_path):

    valid_parameter_source = make_template_parameters_valid(
        request.getfixturevalue(template_parameter_source)
    )

    if isinstance(valid_parameter_source, str):
        param_filename = "params.yml"
        with open(tmp_path/param_filename, 'w') as param_file:
            param_file.write(valid_parameter_source)

        paramset = ParameterSet(param_filepath=tmp_path/param_filename)
        assert param_filename in str(paramset)
        assert 'original=True' in str(paramset)
        return

    if isinstance(valid_parameter_source, dict):
        paramset = ParameterSet(param_dict=valid_parameter_source)
        assert 'dict' in str(paramset)
        assert 'original=True' in str(paramset)
        return

    raise ValueError(
        "Invalid `valid_parameter_source` fixture parametrisation."
    )


# Use `default_parameters` fixture to test the valid parameter set.
def test_ParameterSet___getitem__(valid_paramset, default_parameters):

    # Modify 'verbose' to suppress logging.
    default_parameters.update(verbose=60)

    for key, val in default_parameters.items():
        # This is three tests in one: TypeError, KeyError and value comparison.
        if valid_paramset[key] != val:
            warnings.warn(
                "`valid_paramset` fixture does not match default parameters "
                "for key '{}': {} != {}."
                .format(key, valid_paramset[key], val),
                category=RuntimeWarning
            )


@pytest.mark.parametrize(
    "param_name, param_value",
    [
        ('degrees', {'ell1': 2, 'ell2': 0, 'ELL': 2}),
        ('interlace', True),
    ]
)
def test_ParameterSet___setitem__(param_name, param_value, valid_paramset):

    # Parameter setting should not work when enabling interlacing
    # for three-point statistics.
    if valid_paramset.npoint == '3pt' and param_name == 'interlace':
        valid_paramset[param_name] = param_value
        with pytest.raises(AssertionError):
            assert valid_paramset[param_name] == param_value, \
                "Parameter setting overridden."
        valid_paramset['statistic_type'] = 'powspec'

    # Otherwise, parameter setting should work.
    valid_paramset[param_name] = param_value
    assert valid_paramset[param_name] == param_value, \
        "Parameter set value setting failed."


# Use `default_parameters` fixture to test the valid parameter set.
def test_ParameterSet__getattr__(valid_paramset, default_parameters):

    # Modify 'verbose' to suppress logging.
    default_parameters.update(verbose=60)

    for attr, val in default_parameters.items():
        # This is three tests in one: TypeError, KeyError and value comparison.
        if getattr(valid_paramset, attr) != val:
            warnings.warn(
                "`valid_paramset` fixture does not match default parameters "
                "for attribute '{}': {} != {}."
                .format(attr, getattr(valid_paramset, attr), val),
                category=RuntimeWarning
            )


@pytest.mark.parametrize(
    "param_name, param_value",
    [
        ('degrees', {'ell1': 2, 'ell2': 0, 'ELL': 2}),
        ('interlace', True),
    ]
)
def test_ParameterSet___setattr__(param_name, param_value, valid_paramset):

    # Parameter setting should not work when enabling interlacing
    # for three-point statistics.
    if valid_paramset.npoint == '3pt' and param_name == 'interlace':
        setattr(valid_paramset, param_name, param_value)
        with pytest.raises(AssertionError):
            assert getattr(valid_paramset, param_name) == param_value, \
                "Parameter setting overridden."
        valid_paramset.statistic_type = 'powspec'

    # Otherwise, parameter setting should work.
    setattr(valid_paramset, param_name, param_value)
    assert getattr(valid_paramset, param_name) == param_value, \
        "Parameter set value setting failed."


def test_ParameterSet_names(valid_paramset):
    # There are three derived parameters after validation:
    # 'assignment_order', 'npoint', 'space'.
    assert set(valid_paramset.names()) - set(_TMPL_PARAM_DICT.keys()) == {
        'assignment_order', 'npoint', 'space',
        'fftw_planner_flag', 'fftw_wisdom_file_f', 'fftw_wisdom_file_b'
    }, "Parameter set names do not match internal records."


def test_ParameterSet_items(valid_paramset):
    # There are two derived parameters after validation, 'npoint' and 'space'.
    assert valid_paramset.items(), "Parameter set items not returned."


@pytest.mark.parametrize(
    "param_name, param_value",
    [
        ('npoint', '3pt'),
        ('free_lunch', None),
    ]
)
def test_ParameterSet_get(param_name, param_value, valid_paramset):
    assert valid_paramset.get(param_name) == param_value, \
        "Parameter set parameter not returned correctly."


@pytest.mark.parametrize(
    "update_args, update_kwargs",
    [
        ([dict(verbose=60)], {'fftw_scheme': 'patient'}),
    ]
)
def test_ParameterSet_update(update_args, update_kwargs, valid_paramset):

    valid_paramset.update(*update_args, **update_kwargs)

    for update_arg in update_args:
        for _name, _value in update_arg.items():
            assert valid_paramset.get(_name) == _value, \
                "Parameter set update by positional argument failed."
    for _name, _value in update_kwargs.items():
        assert valid_paramset.get(_name) == _value, \
            "Parameter set update by keyword argument failed."


def test_ParameterSet_print(valid_paramset, capsys):
    valid_paramset.print()
    assert pformat(dict(valid_paramset.items()), sort_dicts=False) \
        in capsys.readouterr().out, "Parameter set misprinted."


def test_ParameterSet_save(valid_paramset, tmp_path):

    param_filepath = tmp_path/"params_saved.yml"

    valid_paramset.save(filepath=param_filepath)

    with open(param_filepath) as param_file:
        valid_param_dict = yaml.load(param_file, Loader=yaml.SafeLoader)

    assert dict(valid_paramset.items()) == valid_param_dict, \
        "Loaded parameter set does not match the saved one."



"""
Data Objects (:mod:`~triumvirate.dataobjs`)
===========================================

Define data objects.

.. autosummary::
    Binning

"""
import numpy as np

from .dataobjs cimport CppBinning


cdef class Binning:
    """Binning.

    Parameters
    ----------
    space : {'config', 'fourier'}
        Coordinate space.
    scheme : {'lin', 'log', 'linpad', 'logpad'}
        Binning scheme.  See the :ref:`note <binning_scheme>` below
        for more details.
    bin_min, bin_max : float, optional
        Minimum and maximum of the bin range (defaults are `None`).
    num_bins : int, optional
        Number of bins (default is `None`).

    Attributes
    ----------
    space : {'config', 'fourier'}
        Coordinate space.
    scheme : {'lin', 'log', 'linpad', 'logpad'}
        Binning scheme.  See the :ref:`note <binning_scheme>` below
        for more details.
    bin_min : float or None
        Minimum of the bin range.
    bin_max : float or None
        Maximum of the bin range.
    num_bins : int or None
        Number of bins.
    bin_edges : list of float
        Bin edges of length (:attr:`num_bins` + 1).
    bin_centres : list of float
        Bin centres of length :attr:`num_bins`.
    bin_widths : list of float
        Bin widths of length :attr:`num_bins`.


    .. _binning_scheme:

    .. admonition:: Binning `scheme`

        The bin setting method supports linear ('lin') and
        log-linear/exponential ('log') binning, with possible linear
        padding from zero for the first 5 bins ('linpad' or 'logpad').
        The padding is either 1.e-3 ('fourier' space) and 10. ('config'
        space), or determined by the mesh grid resolution if set using
        mesh grid sizes.

    """

    def __cinit__(self, space, scheme,
                  bin_min=None, bin_max=None, num_bins=None):

        self.thisptr = new CppBinning(
            space.encode('utf-8'), scheme.encode('utf-8')
        )

        self.scheme = scheme
        self.space = space

        _settable = True
        if bin_min is not None:
            self.bin_min = bin_min
        else:
            _settable = False
        if bin_max is not None:
            self.bin_max = bin_max
        else:
            _settable = False
        if num_bins is not None:
            self.num_bins = num_bins
        else:
            _settable = False

        if _settable:
            self.set_bins(self.bin_min, self.bin_max, self.num_bins)

    def __dealloc__(self):
        del self.thisptr

    @classmethod
    def from_parameter_set(cls, paramset):
        """Create binning from a parameter set.

        This sets :attr:`scheme` and :attr:`space`.

        Parameters
        ----------
        paramset : :class:`~triumvirate.parameters.ParameterSet`
            Parameter set.

        Raises
        ------
        ValueError
            When the 'space' parameter is unset or unrecognised in
            `paramset`.  See the :ref:`note <binning_space>` below
            for more details.


        .. _binning_space:

        .. attention::

            If `paramset` has been initialised without an appropriate
            'statistic_type' parameter value, the derived 'space'
            parameter is unset and must be specified to be the either of
            {'fourier', 'config'} before `paramset` is to this method.
            See also :class:`~triumvirate.parameters.ParameterSet`.

        """
        if paramset['space'] not in ['fourier', 'config']:
            raise ValueError(
                "'space' parameter is unset or unrecognised in `paramset`."
            )

        self = cls(
            paramset['space'], paramset['binning'],
            bin_min=paramset['range'][0],
            bin_max=paramset['range'][-1],
            num_bins=paramset['num_bins']
        )

        return self

    def set_bins(self, bin_min, bin_max, num_bins):
        """Set binning based on the bin range and number using the
        current binning scheme.

        Bin edges, centres and widths are recalculated.

        Parameters
        ----------
        bin_min, bin_max : float
            Minimum and maximum of the bin range.
        num_bins : int
            Number of bins.

        """
        self.bin_min, self.bin_max = bin_min, bin_max
        self.num_bins = num_bins

        self.thisptr.set_bins(self.bin_min, self.bin_max, self.num_bins)

        self.bin_edges = self.thisptr.bin_edges
        self.bin_centres = self.thisptr.bin_centres
        self.bin_widths = self.thisptr.bin_widths

    def set_grid_based_bins(self, boxsize, ngrid):
        """Set linear binning based on a mesh grid.

        The binning :attr:`scheme` is overridden to 'lin'.  The bin width
        is given by the grid resolution in configuration space or the
        fundamental wavenumber in Fourier space.  The bin minimum is zero
        and the bin maximum is half the box size in configuration space or
        the Nyquist wavenumber in Fourier space.

        Parameters
        ----------
        boxsize : float or sequence of [float, float, float]
            Mesh box size (maximum).
        ngrid : int or sequence of [int, int, int]
            Mesh grid number (minimum).

        """
        boxsize = np.asarray(boxsize).max()
        ngrid = np.asarray(ngrid).min()

        self.thisptr.set_bins(boxsize, ngrid)

        self.bin_min, self.bin_max = self.thisptr.bin_min, self.thisptr.bin_max
        self.num_bins = self.thisptr.num_bins
        self.bin_edges = self.thisptr.bin_edges
        self.bin_centres = self.thisptr.bin_centres
        self.bin_widths = self.thisptr.bin_widths
        self.scheme = self.thisptr.scheme

    def set_custom_bins(self, bin_edges):
        """Set customised binning using bin edges.

        The binning :attr:`scheme` is set to 'custom'.
        :attr:`bin_centres` and :attr:`bin_widths` are automatically
        calculated.

        Parameters
        ----------
        bin_edges : 1-d array of float
            Bin edges of length (:attr:`num_bins` + 1).

        """
        self.bin_min, self.bin_max = bin_edges[0], bin_edges[-1]
        self.num_bins = len(bin_edges) - 1

        self.bin_edges = np.array(bin_edges)
        self.bin_centres = np.add(bin_edges[:-1], bin_edges[1:]) / 2.
        self.bin_widths = np.subtract(bin_edges[1:], bin_edges[:-1])

        self.scheme = 'custom'

        self.thisptr.bin_min = self.bin_min
        self.thisptr.bin_max = self.bin_max
        self.thisptr.num_bins = self.num_bins
        self.thisptr.bin_edges = self.bin_edges
        self.thisptr.bin_centres = self.bin_centres
        self.thisptr.bin_widths = self.bin_widths
        self.thisptr.scheme = self.scheme.encode('utf-8')


# NOTE: Remove docstrings from Cython pseudo special functions.
vars()['__reduce_cython__'].__doc__ = None
vars()['__setstate_cython__'].__doc__ = None



"""Test :mod:`~triumvirate.transforms`.

"""
import numpy as np
import pytest
from scipy.special import gamma

from triumvirate._fftlog import HankelTransform
from triumvirate.transforms import (
    DoubleSphericalBesselTransform,
    SphericalBesselTransform,
    resample_lglin,
    resample_lin,
)


@pytest.fixture
def hankel_func_pair(request):
    """Analytical function pair for testing the Hankel transform.

    """
    if request.param == 'sym':
        return (
            lambda r, mu=0., **kwargs: r ** (mu + 1.) * np.exp(-r*r/2.),
            lambda k, mu=0., **kwargs: k ** (mu + 1.) * np.exp(-k*k/2.),
        )
    if request.param == 'asym':
        return (
            lambda r, mu=0., **kwargs: r * (1. + r**2)**(-3./2),
            lambda k, mu=0., **kwargs: k * np.exp(-k),
        )
    raise ValueError("Unknown analytical function pair.")


@pytest.fixture
def sj_func_pair(request):
    """Analytical function pair for testing the spherical
    Bessel transform.

    """
    if request.param == 'sym':
        return (
            lambda r, ell=0, **kwargs:
                r ** ell * np.exp(-r*r/2.),
            lambda k, ell=0, **kwargs:
                k ** ell * np.exp(-k*k/2.) * (2*np.pi)**(3./2),
        )
    if request.param == 'asym':
        return (
            lambda r, s=1, **kwargs: r**(s - 3),
            lambda k, s=1, ell=0, **kwargs:
                np.pi**(3./2)
                * gamma(s/2. + ell/2.) / gamma(3./2 - s/2. + ell/2.)
                * (2./k)**s,
        )


@pytest.mark.parametrize(
    "order, bias, nsamp, lgrange_samp, extrap, hankel_func_pair, lgrange_test",
    [
        (2, 0, 2**10, [-5., 5.], 0, 'sym', [-3., .5]),
        (0, 0, 2**11, [-5., 5.], 4, 'asym', [-3., .5]),
    ],
    indirect=['hankel_func_pair',]
)
def test_hankeltransform(order, bias, nsamp, lgrange_samp, extrap,
                         hankel_func_pair, lgrange_test):

    f, g = hankel_func_pair

    pre_sampts = np.logspace(*lgrange_samp, nsamp, base=10, endpoint=False)
    transformer = HankelTransform(
        order, bias, pre_sampts, kr_c=1., lowring=True, extrap=extrap
    )
    post_sampts = np.asarray(transformer._post_sampts)

    pre_samples = f(pre_sampts, mu=order)
    _, post_samples = transformer.transform(pre_samples)
    post_results = g(post_sampts, mu=order)

    test_range = (lgrange_test[0] < np.log10(post_sampts)) \
        & (np.log10(post_sampts) < lgrange_test[-1])
    assert np.allclose(
        post_samples[test_range].real, post_results[test_range],
        rtol=1.e-3
    ), "Transform not accurate in the expected range."


@pytest.mark.parametrize(
    "degree, bias, nsamp, lgrange_samp, extrap, sj_func_pair, lgrange_test",
    [
        (0, 0, 2**10, [-5., 5.], 0, 'sym', [-3.5, .5]),
        (2, 0, 2**11, [-5., 5.], 4, 'asym', [-1., 3.]),
    ],
    indirect=['sj_func_pair',]
)
def test_sphericalbesseltransform(degree, bias, nsamp, lgrange_samp, extrap,
                                  sj_func_pair, lgrange_test):

    f, g = sj_func_pair

    pre_sampts = np.logspace(*lgrange_samp, nsamp, base=10, endpoint=False)
    pre_samples = f(pre_sampts, ell=degree)

    transformer = SphericalBesselTransform(
        degree, bias, pre_sampts, lowring=True, extrap=extrap
    )

    # Test only
    # :meth:`~triumvirate.transforms.SphericalBesselTransform.transform`
    # as a proxy for
    # :class:`~triumvirate.transforms.SphericalBesselTransform.transform_cosmo_multipoles`.
    post_sampts, post_samples = transformer.transform(pre_samples)
    post_results = g(post_sampts, ell=degree)

    test_range = (lgrange_test[0] < np.log10(post_sampts)) \
        & (np.log10(post_sampts) < lgrange_test[-1])
    assert np.allclose(
        post_samples[test_range].real, post_results[test_range],
        rtol=1.e-3
    ), "Transform not accurate in the expected range."


@pytest.mark.parametrize(
    "degrees, biases, nsamp, lgrange_samp, extrap, sj_func_pair, lgrange_test",
    [
        ((0, 0), (0, 0), 2**10, [-5., 5.], 0, 'sym', [-3., 0.]),
        ((2, 2), (0, 0), 2**11, [-5., 5.], 4, 'asym', [-.5, 3.]),
    ],
    indirect=['sj_func_pair',]
)
def test_doublesphericalbesseltransform(degrees, biases, nsamp, lgrange_samp,
                                        extrap, sj_func_pair, lgrange_test):

    f, g = sj_func_pair

    pre_sampts = np.logspace(*lgrange_samp, nsamp, base=10, endpoint=False)
    pre_samples_1, pre_samples_2 = (
        f(pre_sampts, ell=ell) for ell in degrees
    )
    pre_samples = pre_samples_1[:, None] * pre_samples_2[None, :]

    transformer = DoubleSphericalBesselTransform(
        degrees, biases, pre_sampts, lowring=True, extrap=extrap
    )

    # Test only
    # :meth:`~triumvirate.transforms.DoubleSphericalBesselTransform.transform`
    # as a proxy for
    # :class:`~triumvirate.transforms.DoubleSphericalBesselTransform.transform_cosmo_multipoles`.
    post_sampts, post_samples = transformer.transform(pre_samples)
    post_results = np.multiply(*[
        g(post_sampts_i, ell=ell)
        for (post_sampts_i, ell) in zip(post_sampts, degrees)
    ])

    test_range = np.logical_and.reduce([
        np.log10(post_sampts[0]) > lgrange_test[0],
        np.log10(post_sampts[0]) < lgrange_test[-1],
        np.log10(post_sampts[-1]) > lgrange_test[0],
        np.log10(post_sampts[-1]) < lgrange_test[-1],
    ])
    assert np.allclose(
        post_samples[test_range].real, post_results[test_range],
        rtol=1.e-3
    ), "Transform not accurate in the expected range."


@pytest.mark.parametrize(
    "sampts, size, func, ndim",
    [
        (np.arange(1., 11.), None, lambda x: x**2, 1),
        (
            np.arange(1., 51.) * 1.e-3, 10,
            lambda x, y: np.exp(-x*np.sqrt(y)),
            2
        ),
    ]
)
def test_resample_lglin(sampts, size, func, ndim):

    if ndim == 1:
        samples = func(sampts)
    elif ndim == 2:
        samples = func(*np.meshgrid(sampts, sampts, indexing='ij'))
    else:
        raise ValueError("Resampling only supports 1- or 2-d arrays.")

    resampts, resamples = resample_lglin(sampts, samples, size=size)

    if ndim == 1:
        resamples_exact = func(resampts)
    if ndim == 2:
        resamples_exact = func(
            *np.meshgrid(resampts[0], resampts[-1], indexing='ij')
        )

    assert np.allclose(resamples, resamples_exact), "Resampling not accurate."


@pytest.mark.parametrize(
    "sampts, size, func, ndim",
    [
        (np.arange(1., 11.), None, lambda x: x**2, 1),
        (np.arange(1., 51.) * 1.e-3, 10, lambda x, y: x + y, 2),
    ]
)
def test_resample_lin(sampts, size, func, ndim):

    if ndim == 1:
        samples = func(sampts)
    elif ndim == 2:
        samples = func(*np.meshgrid(sampts, sampts, indexing='ij'))
    else:
        raise ValueError("Resampling only supports 1- or 2-d arrays.")

    resampts, resamples = resample_lin(sampts, samples, size=size)

    if ndim == 1:
        resamples_exact = func(resampts)
    if ndim == 2:
        resamples_exact = func(
            *np.meshgrid(resampts[0], resampts[-1], indexing='ij')
        )

    assert np.allclose(resamples, resamples_exact), "Resampling not accurate."



"""Interface with the FFTLog algorithms for the Hankel and
associated transforms.

"""
from libcpp cimport bool as bool_t
from libcpp.vector cimport vector


cdef extern from "include/fftlog.hpp":
    cdef cppclass CppHankelTransform "trv::maths::HankelTransform":
        double order
        double bias
        int nsamp
        int nsamp_trans
        double logres
        double pivot

        vector[double] pre_sampts
        vector[double] post_sampts

        CppHankelTransform(double mu, double q, bool_t threaded)

        void initialise(
            vector[double] sample_pts, double kr_c, bool_t lowring,
            int extrap, double extrap_exp
        ) except +

        void biased_transform(
            double complex* a, double complex* b
        ) except +


cdef class HankelTransform:
    cdef CppHankelTransform* thisptr
    cdef public double order
    cdef public double bias
    cdef public int _nsamp
    cdef public int _nsamp_trans
    cdef public double _logres
    cdef public double _pivot
    cdef public vector[double] _pre_sampts
    cdef public vector[double] _post_sampts



"""
Installation Validation (:mod:`~triumvirate._valid_install`)
============================================================

Check whether the Cython extensions have been compiled correctly by
performing test computations.

It is not meant to replace the full `pytest` suite, and does not check the
correctness of the computed results.

"""  # numpydoc ignore=SS01
import warnings

import numpy as np

from .catalogue import ParticleCatalogue
from .dataobjs import Binning
from .parameters import ParameterSet, fetch_paramset_template
from .twopt import compute_powspec_in_gpp_box


def _create_minimal_catalogue():
    """Create a minimal catalogue of a single particle at a unit length
    away from all positive axes.

    Returns
    -------
    :class:`~triumvirate.catalogue.ParticleCatalogue`
        Minimal catalogue.

    """
    CTLG_LENGTH = 1

    x = y = z = np.ones(CTLG_LENGTH)

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message=".*'nz' field is None.*")
        return ParticleCatalogue(x, y, z)


def _define_minimal_binning():
    """Define a minimal binning scheme.

    Returns
    -------
    :class:`~triumvirate.dataobjs.Binning`
        Minimal binning.

    """
    SPACE, SCHEME = 'fourier', 'lin'
    BIN_MIN, BIN_MAX, NUM_BINS = 0., .1, 2

    return Binning(
        SPACE, SCHEME, bin_min=BIN_MIN, bin_max=BIN_MAX, num_bins=NUM_BINS
    )


def _create_mock_paramset():
    """Create a mock parameter set.

    Returns
    -------
    :class:`~triumvirate.parameters.ParameterSet`
        Mock parameter set.

    """
    tmpl_paramdict = fetch_paramset_template('dict')

    for ax_name in ['x', 'y', 'z']:
        tmpl_paramdict['boxsize'][ax_name] = 2.
        tmpl_paramdict['ngrid'][ax_name] = 2.

    tmpl_paramdict.update({
        'degrees': {'ell1': 0, 'ell2': 0, 'ELL': 0},
        'range': [0., 1.],
        'num_bins': 2,
        'verbose': 60,
    })

    return ParameterSet(param_dict=tmpl_paramdict)


def validate_installation():
    """Validate installation by computing a minimal test case.

    Returns
    -------
    True
        If the installation is valid.

    """
    catalogue = _create_minimal_catalogue()
    binning = _define_minimal_binning()
    paramset = _create_mock_paramset()

    compute_powspec_in_gpp_box(catalogue, binning=binning, paramset=paramset)

    print("Installation of Triumvirate has been validated.")

    return True



"""
Field and Mesh (:mod:`~triumvirate.fieldmesh`)
==============================================

.. versionadded:: 0.3.0


Handle fields and mesh grids.

.. autosummary::
    record_binned_vectors

"""  # numpydoc ignore=SS01
import numpy as np

from ._field import _record_binned_vectors
from .parameters import (
    _modify_sampling_parameters,
    fetch_paramset_template,
    ParameterSet,
)


def record_binned_vectors(binning, paramset=None, boxsize=None, ngrid=None):
    """Record binned vectors given a binning scheme and mesh grid
    parameters.

    Parameters
    ----------
    binning : :class:`~triumvirate.binning.Binning`
        Binning.
    paramset : :class:`~triumvirate.parameters.ParameterSet`, optional
        Parameters set including the mesh grid parameters.  If `None`
        (default), `boxsize` and `ngrid` must be provided.
    boxsize : float or sequence of [float, float, float], optional
        Box size in each dimension (default is `None`).  Must be provided
        if `paramset` is `None`.  This will override the corresponding
        entries in `paramset`.
    ngrid : int or sequence of [int, int, int], optional
        Grid number in each dimension (default is `None`).  Must be
        provided if `paramset` is set.  This will override the
        corresponding entries in `paramset`.

    Returns
    -------
    binned_vectors : :class:`numpy.ndarray`
        Binned vectors as a structured array with the following fields---

        - ``'index'``: bin index;
        - ``'lower_edge'``: lower edge of the bin;
        - ``'upper_edge'``: upper edge of the bin;
        - ``'vecx'``: x-component of the vector;
        - ``'vecy'``: y-component of the vector;
        - ``'vecz'``: z-component of the vector,

    """
    if paramset is None:
        paramset = fetch_paramset_template('dict')

        paramset.update(range=[binning.bin_min, binning.bin_max])  # filler
        paramset.update(num_bins=binning.num_bins)  # filler

        paramset.update(
            degrees={'ell1': 0, 'ell2': 0, 'ELL': 0}
        )  # placeholder

        if boxsize is None or ngrid is None:
            raise ValueError(
                "Either `paramset` or both `boxsize` and `ngrid` "
                "must be provided."
            )

    params_sampling = {}
    if boxsize is not None:
        params_sampling['boxsize'] = [boxsize] * 3 if np.isscalar(boxsize) \
            else list(boxsize)
    if ngrid is not None:
        params_sampling['ngrid'] = [ngrid] * 3 if np.isscalar(ngrid) \
            else list(ngrid)

    if params_sampling:
        paramset = _modify_sampling_parameters(paramset, params_sampling)

    if isinstance(paramset, dict):  # likely redundant but safe
        paramset = ParameterSet(param_dict=paramset)

    binned_vectors = _record_binned_vectors(binning, paramset)

    return binned_vectors



"""Test :mod:`~triumvirate.winconv`.

"""
import doctest
from contextlib import nullcontext as does_not_raise

import numpy as np
import pytest
from sympy import Expr, sympify

import triumvirate.winconv as winconv
from triumvirate.winconv import (
    # BispecWinConv,
    # PowspecWinConv,
    # ThreePCFWinConv,
    # ThreePointWindow,
    # TwoPCFWinConv,
    WinConvBase,
    # calc_threept_ic,
    calc_threept_winconv_coeff,
)


def test_doctest():
    # Doctest is included for `Multipole`, `WinConvTerm` and `WinConvFormulae`.
    assert doctest.testmod(winconv, raise_on_error=True).failed == 0


@pytest.mark.parametrize(
    "multipole, multipole_Q, multipole_Z, expected_coeff",
    [
        ('000', '110', '110', sympify('1/3')),
        ('110', '000', '110', 1),
        ('110', '000', '000', 0.),
    ]
)
def test_calc_threept_winconv_coeff(multipole, multipole_Q, multipole_Z,
                                    expected_coeff):

    coeff = calc_threept_winconv_coeff(
        multipole, multipole_Q, multipole_Z,
        symb=isinstance(expected_coeff, Expr)
    )

    if isinstance(expected_coeff, Expr):
        validity = (coeff == expected_coeff)
    else:
        validity = np.isclose(coeff, expected_coeff)
    assert validity


@pytest.mark.parametrize(
    "formulae, window_sampts, window_multipoles, exception_context",
    [
        (
            "wilson+16",
            np.array([0.1, 0.2, 0.3]),
            {
                '0': np.array([1.0, 2.0, 3.0]),
                '2': np.array([0.5, 0.6, 0.7]),
            },
            pytest.raises(ValueError),  # missing multipole
        ),
        (
            "wilson+16",
            np.array([0.1, 0.2, 0.3]),
            {
                '0': np.array([1.0, 2.0, 3.0]),
                '2': np.array([0.5, 0.6, 0.7]),
                '4': np.array([0.5, 0.6, 0.7]),
                '6': np.array([0.5, 0.6, 0.7]),
                '8': np.array([0.5, 0.6, 0.7]),
            },
            does_not_raise(),
        ),
        (
            "sugiyama+18",
            np.array([0.1, 0.2, 0.3]),
            {
                '000': np.array(
                    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
                ),
                '202': np.array(
                    [[0.5, 0.6, 0.7], [0.8, 0.9, 1.0], [1.1, 1.2, 1.3]]
                ),
                '110': np.array(
                    [[0.5, 0.6, 0.7], [0.8, 0.9, 1.0], [1.1, 1.2, 1.3]]
                ),
                '112': np.array(
                    [[0.5, 0.6, 0.7], [0.8, 0.9, 1.0], [1.1, 1.2, 1.3]]
                ),
                '022': np.array(
                    [[0.5, 0.6, 0.7], [0.8, 0.9, 1.0], [1.1, 1.2, 1.3]]
                ),
            },
            does_not_raise(),
        ),
        (
            "sugiyama+18",
            np.array([0.1, 0.2, 0.3]),
            {
                '000': np.array(
                    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
                ),
                '202': np.array(
                    [[0.5, 0.6, 0.7], [0.8, 0.9, 1.0], [1.1, 1.2, 1.3]]
                ),
            },
            pytest.raises(ValueError),  # missing multipole
        ),
    ]
)
def test_winconvbase(formulae, window_sampts, window_multipoles,
                     exception_context):

    with exception_context:
        _ = WinConvBase(formulae, window_sampts, window_multipoles)



"""
FFTLog algorithm (:mod:`~triumvirate._fftlog`)
==============================================

Implement the FFTLog algorithm for Hankel-related transforms.

"""
import numpy as np
cimport numpy as np

from ._arrayops import _check_1d_array
from ._fftlog cimport CppHankelTransform

np.import_array()


cdef class HankelTransform:
    """Hankel transform.

    Parameters
    ----------
    mu : float
        Order of the Hankel transform.
    q : float
        Power-law bias index.
    x : array of float
        Pre-transform sample points.  Must be logarithmically spaced.
        Must be even in length if extrapolation is used.
    kr_c : float
        Pivot value for the transform.  When `lowring` is `True`, this is
        adjusted if it is non-zero, or otherwise directly calculated.
    lowring : bool, optional
        Low-ringing condition (default is `True`).
    extrap : int, optional
        Extrapolation method (default is 0) with the following
        options:

        - 0: none;
        - 1: extrapolate by constant padding;
        - 2: extrapolate linearly;
        - 3: extrapolate log-linearly.

        Any extrapolation results in a sample size for the transform that
        is the smallest power of 2 greater than or equal to `extrap_exp`
        times the original number of sample points; the pre-transform
        samples are assumed to be real.
    extrap_exp : float, optional
        Sample size expansion factor (default is 2.) for extrapolation.
        The smallest power of 2 greater than or equal to this times the
        original number of sample points is used as the sample size for
        the transform.
    threaded : bool, optional
        If `True` (default is `False`), use the multi-threaded FFTLog
        algorithm.

    Attributes
    ----------
    order : float
        Order of the transform.
    bias : float
        Power-law bias index.
    size : int
        Sample size of the transform.
    pivot : float
        Pivot value used.

    """

    def __cinit__(self, mu, q, x, kr_c, lowring=True, extrap=0, extrap_exp=2.,
                  threaded=False):

        _x = np.ascontiguousarray(_check_1d_array(x, check_loglin=True).copy())

        self.thisptr = new CppHankelTransform(<double>mu, <double>q, threaded)

        self.thisptr.initialise(
            _x, kr_c, lowring, 0 if extrap is None else extrap, extrap_exp
        )

        self.order = self.thisptr.order
        self.bias = self.thisptr.bias

        self._nsamp = self.thisptr.nsamp
        self._nsamp_trans = self.thisptr.nsamp_trans
        self._logres = self.thisptr.logres
        self._pivot = self.thisptr.pivot

        self._pre_sampts = self.thisptr.pre_sampts
        self._post_sampts = self.thisptr.post_sampts

    def __dealloc__(self):
        del self.thisptr

    @property
    def size(self):
        """Sample size.

        """
        return self._nsamp_trans

    @property
    def pivot(self):
        """Pivot value.

        """
        return self._pivot

    def transform(self, fx):
        """Transform samples at initialised sample points.

        Parameters
        ----------
        fx : array_like
            Pre-transform samples.  Must be even in length, and assumed
            to be effectively real, if extrapolation is used.

        Returns
        -------
        y, gy : array_like
            Post-transform sample points and samples.

        """
        fx_ = fx.astype(np.complex128)
        gy_ = np.zeros_like(fx, dtype=np.complex128)
        self._transform(fx_, gy_)

        y = np.array(self._post_sampts)
        gy = np.array(gy_)

        return y, gy

    def _transform(self,
                   np.ndarray[double complex, ndim=1, mode='c'] fx not None,
                   np.ndarray[double complex, ndim=1, mode='c'] gy not None):
        """Perform underlying sample transforms.

        Parameters
        ----------
        fx : array_like
            Pre-transform samples.
        gy : array_like
            Post-transform samples.

        """
        self.thisptr.biased_transform(&fx[0], &gy[0])



"""Interface with the parameter set.

"""
from libcpp cimport bool as bool_t
from libcpp.string cimport string


cdef extern from "include/parameters.hpp":
    cdef cppclass CppParameterSet "trv::ParameterSet":
        # ----------------------------------------------------------------
        # Members
        # ----------------------------------------------------------------

        # -- I/O ---------------------------------------------------------

        string catalogue_dir
        string measurement_dir
        string data_catalogue_file
        string rand_catalogue_file
        # string catalogue_columns
        string output_tag

        # -- Mesh sampling -----------------------------------------------

        double boxsize[3]
        int ngrid[3]

        double volume
        long long nmesh

        string alignment
        string padscale
        double padfactor

        string assignment
        string interlace
        int assignment_order

        # -- Measurement -------------------------------------------------

        string catalogue_type
        string statistic_type
        string npoint
        string space

        string form
        string norm_convention

        string binning

        int ell1
        int ell2
        int ELL

        int i_wa
        int j_wa

        double bin_min
        double bin_max

        int num_bins
        int idx_bin

        # -- Misc --------------------------------------------------------

        string fftw_scheme
        unsigned fftw_planner_flag
        string use_fftw_wisdom
        string fftw_wisdom_file_f
        string fftw_wisdom_file_b
        # string save_binned_vectors
        int verbose

        # ----------------------------------------------------------------
        # Methods
        # ----------------------------------------------------------------

        int validate() except +

        # int print_to_file(char* out_parameter_filepath)
        # int print_to_file()


cdef class ParameterSet:
    cdef CppParameterSet* thisptr
    cdef public object _logger
    cdef public str _source
    cdef public bool_t _original
    cdef public bool_t _validity
    cdef public dict _params
